@@ -51,7 +51,6 @@ static int list_dir_proc(ext2_ino_t dir,
 			 char	*buf,
 			 void	*private)
 {
-	struct ext2_dir_entry_2	*d2;
 	struct ext2_inode	inode;
 	ext2_ino_t		ino;
 	struct tm		*tm_p;
@@ -90,9 +89,9 @@ static int list_dir_proc(ext2_ino_t dir,
 			strcpy(datestr, "                 ");
 			memset(&inode, 0, sizeof(struct ext2_inode));
 		}
-		d2 = (struct ext2_dir_entry_2 *) dirent;
 		fprintf(ls->f, "%c%6u%c %6o (%d)  %5d  %5d   ", lbr, ino, rbr,
-			inode.i_mode, d2->file_type, inode.i_uid, inode.i_gid);
+			inode.i_mode, dirent->name_len >> 8,
+			inode.i_uid, inode.i_gid);
 		if (LINUX_S_ISDIR(inode.i_mode))
 			fprintf(ls->f, "%5d", inode.i_size);
 		else
