@@ -58,6 +58,7 @@
 #define FSCK_UNCORRECTED 4	/* File system errors left uncorrected */
 #define FSCK_ERROR       8	/* Operational error */
 #define FSCK_USAGE       16	/* Usage or syntax error */
+#define FSCK_CANCELED	 32	/* Aborted with a signal or ^C */
 #define FSCK_LIBRARY     128	/* Shared library error */
 
 /*
@@ -77,6 +78,41 @@ struct dir_info {
 	ext2_ino_t		parent; /* Parent according to treewalk */
 };
 
+
+/*
+ * The indexed directory information structure; stores information for
+ * directories which contain a hash tree index.
+ */
+struct dx_dir_info {
+	ext2_ino_t		ino; 		/* Inode number */
+	int			numblocks;	/* number of blocks */
+	int			hashversion;
+	short			depth;		/* depth of tree */
+	struct dx_dirblock_info	*dx_block; 	/* Array of size numblocks */
+};
+
+#define DX_DIRBLOCK_ROOT	1
+#define DX_DIRBLOCK_LEAF	2
+#define DX_DIRBLOCK_NODE	3
+#define DX_DIRBLOCK_CORRUPT	4
+#define DX_DIRBLOCK_CLEARED	8
+
+struct dx_dirblock_info {
+	int		type;
+	blk_t		phys;
+	int		flags;
+	blk_t		parent;
+	ext2_dirhash_t	min_hash; 
+	ext2_dirhash_t	max_hash;
+	ext2_dirhash_t	node_min_hash; 
+	ext2_dirhash_t	node_max_hash;
+};
+
+#define DX_FLAG_REFERENCED	1
+#define DX_FLAG_DUP_REF		2
+#define DX_FLAG_FIRST		4
+#define DX_FLAG_LAST		8
+
 #ifdef RESOURCE_TRACK
 /*
  * This structure is used for keeping track of how much resources have
@@ -103,6 +139,7 @@ struct resource_track {
 #define E2F_OPT_DEBUG		0x0080
 #define E2F_OPT_FORCE		0x0100
 #define E2F_OPT_WRITECHECK	0x0200
+#define E2F_OPT_COMPRESS_DIRS	0x0400
 
 /*
  * E2fsck flags
@@ -151,6 +188,7 @@ struct e2fsck_struct {
 	blk_t	superblock;	/* sb used to open fs */
 	int	blocksize;	/* blocksize */
 	blk_t	num_blocks;	/* Total number of blocks */
+	int	mount_flags;
 
 #ifdef HAVE_SETJMP_H
 	jmp_buf	abort_loc;
@@ -200,6 +238,12 @@ struct e2fsck_struct {
 	ext2_ino_t stashed_ino;
 	struct ext2_inode *stashed_inode;
 
+	/*
+	 * Location of the lost and found directory
+	 */
+	ext2_ino_t lost_and_found;
+	int bad_lost_and_found;
+	
 	/*
 	 * Directory information
 	 */
@@ -207,6 +251,18 @@ struct e2fsck_struct {
 	int		dir_info_size;
 	struct dir_info	*dir_info;
 
+	/*
+	 * Indexed directory information
+	 */
+	int		dx_dir_info_count;
+	int		dx_dir_info_size;
+	struct dx_dir_info *dx_dir_info;
+
+	/*
+	 * Directories to hash
+	 */
+	ext2_u32_list	dirs_to_hash;
+
 	/*
 	 * Tuning parameters
 	 */
@@ -254,6 +310,8 @@ struct e2fsck_struct {
 	int fs_ext_attr_inodes;
 	int fs_ext_attr_blocks;
 
+	int ext_attr_ver;
+
 	/*
 	 * For the use of callers of the e2fsck functions; not used by
 	 * e2fsck functions themselves.
@@ -292,10 +350,16 @@ extern void test_disk(e2fsck_t ctx);
 extern void e2fsck_add_dir_info(e2fsck_t ctx, ext2_ino_t ino, ext2_ino_t parent);
 extern struct dir_info *e2fsck_get_dir_info(e2fsck_t ctx, ext2_ino_t ino);
 extern void e2fsck_free_dir_info(e2fsck_t ctx);
-extern int e2fsck_get_num_dirs(e2fsck_t ctx);
 extern int e2fsck_get_num_dirinfo(e2fsck_t ctx);
 extern struct dir_info *e2fsck_dir_info_iter(e2fsck_t ctx, int *control);
 
+/* dx_dirinfo.c */
+extern void e2fsck_add_dx_dir(e2fsck_t ctx, ext2_ino_t ino, int num_blocks);
+extern struct dx_dir_info *e2fsck_get_dx_dir_info(e2fsck_t ctx, ext2_ino_t ino);
+extern void e2fsck_free_dx_dir_info(e2fsck_t ctx);
+extern int e2fsck_get_num_dx_dirinfo(e2fsck_t ctx);
+extern struct dx_dir_info *e2fsck_dx_dir_info_iter(e2fsck_t ctx, int *control);
+
 /* ea_refcount.c */
 extern errcode_t ea_refcount_create(int size, ext2_refcount_t *ret);
 extern void ea_refcount_free(ext2_refcount_t refcount);
@@ -322,20 +386,31 @@ extern void e2fsck_move_ext3_journal(e2fsck_t ctx);
 
 /* pass1.c */
 extern void e2fsck_use_inode_shortcuts(e2fsck_t ctx, int bool);
-extern int e2fsck_pass1_check_device_inode(struct ext2_inode *inode);
-extern int e2fsck_pass1_check_symlink(ext2_filsys fs, struct ext2_inode *inode);
+extern int e2fsck_pass1_check_device_inode(ext2_filsys fs,
+					   struct ext2_inode *inode);
+extern int e2fsck_pass1_check_symlink(ext2_filsys fs,
+				      struct ext2_inode *inode, char *buf);
 
 /* pass2.c */
-extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir, ext2_ino_t ino);
+extern int e2fsck_process_bad_inode(e2fsck_t ctx, ext2_ino_t dir,
+				    ext2_ino_t ino, char *buf);
 
 /* pass3.c */
 extern int e2fsck_reconnect_file(e2fsck_t ctx, ext2_ino_t inode);
+extern errcode_t e2fsck_expand_directory(e2fsck_t ctx, ext2_ino_t dir,
+					 int num, int gauranteed_size);
+extern ext2_ino_t e2fsck_get_lost_and_found(e2fsck_t ctx, int fix);
+
 
 /* region.c */
 extern region_t region_create(region_addr_t min, region_addr_t max);
 extern void region_free(region_t region);
 extern int region_allocate(region_t region, region_addr_t start, int n);
 
+/* rehash.c */
+errcode_t e2fsck_rehash_dir(e2fsck_t ctx, ext2_ino_t ino);
+void e2fsck_rehash_directories(e2fsck_t ctx);
+
 /* super.c */
 void check_super_block(e2fsck_t ctx);
 errcode_t e2fsck_get_device_size(e2fsck_t ctx);
