@@ -37,6 +37,7 @@
  * found.
  */
 
+#include <string.h>
 #include <time.h>
 #ifdef HAVE_ERRNO_H
 #include <errno.h>
@@ -76,9 +77,10 @@ static void adjust_extattr_refcount(e2fsck_t ctx, ext2_refcount_t refcount,
 
 struct process_block_struct {
 	ext2_ino_t	ino;
-	int		is_dir:1, clear:1, suppress:1,
+	int		is_dir:1, is_reg:1, clear:1, suppress:1,
 				fragmented:1, compressed:1;
 	blk_t		num_blocks;
+	blk_t		max_blocks;
 	e2_blkcnt_t	last_block;
 	int		num_illegal_blocks;
 	blk_t		previous_block;
@@ -103,7 +105,8 @@ struct scan_callback_struct {
 static struct process_inode_block *inodes_to_process;
 static int process_inode_count;
 
-static __u64 ext2_max_sizes[4];
+static __u64 ext2_max_sizes[EXT2_MAX_BLOCK_LOG_SIZE -
+			    EXT2_MIN_BLOCK_LOG_SIZE + 1];
 
 /*
  * Free all memory allocated by pass1 in preparation for restarting
@@ -123,15 +126,18 @@ static void unwind_pass1(ext2_filsys fs)
  * since they have the same requirement; the i_block fields should be
  * zero. 
  */
-int e2fsck_pass1_check_device_inode(struct ext2_inode *inode)
+int e2fsck_pass1_check_device_inode(ext2_filsys fs, struct ext2_inode *inode)
 {
 	int	i;
 
 	/*
-	 * If i_blocks is non-zero, then this is a bogus device/fifo/socket
+	 * If i_blocks is non-zero, or the index flag is set, then
+	 * this is a bogus device/fifo/socket
 	 */
-	if (inode->i_blocks)
+	if ((ext2fs_inode_data_blocks(fs, inode) != 0) ||
+	    (inode->i_flags & EXT2_INDEX_FL))
 		return 0;
+
 	/*
 	 * We should be able to do the test below all the time, but
 	 * because the kernel doesn't forcibly clear the device
@@ -170,18 +176,21 @@ static int strnlen(const char * s, int count)
  * Check to make sure a symlink inode is real.  Returns 1 if the symlink
  * checks out, 0 if not.
  */
-int e2fsck_pass1_check_symlink(ext2_filsys fs, struct ext2_inode *inode)
+int e2fsck_pass1_check_symlink(ext2_filsys fs, struct ext2_inode *inode,
+			       char *buf)
 {
+	int len;
 	int i;
+	blk_t	blocks;
 
 	if ((inode->i_size_high || inode->i_size == 0) ||
-	    (inode->i_flags & (EXT2_IMMUTABLE_FL | EXT2_APPEND_FL |
-			       EXT2_INDEX_FL)))
+	    (inode->i_flags & EXT2_INDEX_FL))
 		return 0;
 
-	if (inode->i_blocks) {
-		if ((inode->i_size > fs->blocksize) ||
-		    (inode->i_blocks != fs->blocksize >> 9) ||
+	blocks = ext2fs_inode_data_blocks(fs, inode);
+	if (blocks) {
+		if ((inode->i_size >= fs->blocksize) ||
+		    (blocks != fs->blocksize >> 9) ||
 		    (inode->i_block[0] < fs->super->s_first_data_block) ||
 		    (inode->i_block[0] >= fs->super->s_blocks_count))
 			return 0;
@@ -189,14 +198,23 @@ int e2fsck_pass1_check_symlink(ext2_filsys fs, struct ext2_inode *inode)
 		for (i = 1; i < EXT2_N_BLOCKS; i++)
 			if (inode->i_block[i])
 				return 0;
-	} else {
-		if (inode->i_size > sizeof(inode->i_block) - 1)
+
+		if (io_channel_read_blk(fs->io, inode->i_block[0], 1, buf))
 			return 0;
 
-		if (inode->i_size !=
-		    strnlen((char *)inode->i_block, sizeof(inode->i_block)))
+		len = strnlen(buf, fs->blocksize);
+		if (len == fs->blocksize)
+			return 0;
+	} else {
+		if (inode->i_size >= sizeof(inode->i_block))
+			return 0;
+
+		len = strnlen((char *)inode->i_block, sizeof(inode->i_block));
+		if (len == sizeof(inode->i_block))
 			return 0;
 	}
+	if (len != inode->i_size)
+		return 0;
 	return 1;
 }
 
@@ -204,15 +222,16 @@ int e2fsck_pass1_check_symlink(ext2_filsys fs, struct ext2_inode *inode)
  * If the immutable (or append-only) flag is set on the inode, offer
  * to clear it.
  */
+#define BAD_SPECIAL_FLAGS (EXT2_IMMUTABLE_FL | EXT2_APPEND_FL)
 static void check_immutable(e2fsck_t ctx, struct problem_context *pctx)
 {
-	if (!(pctx->inode->i_flags & (EXT2_IMMUTABLE_FL | EXT2_APPEND_FL)))
+	if (!(pctx->inode->i_flags & BAD_SPECIAL_FLAGS))
 		return;
 
 	if (!fix_problem(ctx, PR_1_SET_IMMUTABLE, pctx))
 		return;
 
-	pctx->inode->i_flags &= ~((EXT2_IMMUTABLE_FL | EXT2_APPEND_FL));
+	pctx->inode->i_flags &= ~BAD_SPECIAL_FLAGS;
 	e2fsck_write_inode(ctx, pctx->ino, pctx->inode, "pass1");
 }
 
@@ -239,7 +258,7 @@ static void check_size(e2fsck_t ctx, struct problem_context *pctx)
 void e2fsck_pass1(e2fsck_t ctx)
 {
 	int	i;
-	__u64	max_sizes, max_sect_limit;
+	__u64	max_sizes;
 	ext2_filsys fs = ctx->fs;
 	ext2_ino_t	ino;
 	struct ext2_inode inode;
@@ -253,6 +272,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 	struct		scan_callback_struct scan_struct;
 	struct ext2_super_block *sb = ctx->fs->super;
 	int		imagic_fs;
+	int		busted_fs_time = 0;
 	
 #ifdef RESOURCE_TRACK
 	init_resource_track(&rtrack);
@@ -262,23 +282,24 @@ void e2fsck_pass1(e2fsck_t ctx)
 	if (!(ctx->options & E2F_OPT_PREEN))
 		fix_problem(ctx, PR_1_PASS_HEADER, &pctx);
 
+	if ((fs->super->s_feature_compat & EXT2_FEATURE_COMPAT_DIR_INDEX) &&
+	    !(ctx->options & E2F_OPT_NO)) {
+		if (ext2fs_u32_list_create(&ctx->dirs_to_hash, 50))
+			ctx->dirs_to_hash = 0;
+	}
+
 #ifdef MTRACE
 	mtrace_print("Pass 1");
 #endif
 
-#define EXT2_BPP(bits) (1UL << ((bits) - 2))
+#define EXT2_BPP(bits) (1ULL << ((bits) - 2))
 
-	for (i=0; i < 4; i++) {
-		max_sizes = EXT2_NDIR_BLOCKS + EXT2_BPP(10+i);
-		max_sizes = max_sizes + EXT2_BPP(10+i) * EXT2_BPP(10+i);
-		max_sizes = (max_sizes +
-			     (__u64) EXT2_BPP(10+i) * EXT2_BPP(10+i) *
-			     EXT2_BPP(10+i));
-		max_sizes = (max_sizes * (1UL << (10+i))) - 1;
-		max_sect_limit = 512ULL * ((1LL << 32) - (1 << (i+1)));
-		if (max_sizes > max_sect_limit)
-			max_sizes = max_sect_limit;
-		ext2_max_sizes[i] = max_sizes;
+	for (i = EXT2_MIN_BLOCK_LOG_SIZE; i <= EXT2_MAX_BLOCK_LOG_SIZE; i++) {
+		max_sizes = EXT2_NDIR_BLOCKS + EXT2_BPP(i);
+		max_sizes = max_sizes + EXT2_BPP(i) * EXT2_BPP(i);
+		max_sizes = max_sizes + EXT2_BPP(i) * EXT2_BPP(i) * EXT2_BPP(i);
+		max_sizes = (max_sizes * (1UL << i)) - 1;
+		ext2_max_sizes[i - EXT2_MIN_BLOCK_LOG_SIZE] = max_sizes;
 	}
 #undef EXT2_BPP
 
@@ -367,12 +388,6 @@ void e2fsck_pass1(e2fsck_t ctx)
 		return;
 	}
 	ext2fs_inode_scan_flags(scan, EXT2_SF_SKIP_MISSING_ITABLE, 0);
-	pctx.errcode = ext2fs_get_next_inode(scan, &ino, &inode);
-	if (pctx.errcode) {
-		fix_problem(ctx, PR_1_ISCAN_ERROR, &pctx);
-		ctx->flags |= E2F_FLAG_ABORT;
-		return;
-	}
 	ctx->stashed_inode = &inode;
 	scan_struct.ctx = ctx;
 	scan_struct.block_buf = block_buf;
@@ -380,7 +395,27 @@ void e2fsck_pass1(e2fsck_t ctx)
 	if (ctx->progress)
 		if ((ctx->progress)(ctx, 1, 0, ctx->fs->group_desc_count))
 			return;
-	while (ino) {
+	if (fs->super->s_wtime < fs->super->s_inodes_count)
+		busted_fs_time = 1;
+
+	while (1) {
+		pctx.errcode = ext2fs_get_next_inode(scan, &ino, &inode);
+		if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
+			return;
+		if (pctx.errcode == EXT2_ET_BAD_BLOCK_IN_INODE_TABLE) {
+			if (!ctx->inode_bb_map)
+				alloc_bb_map(ctx);
+			ext2fs_mark_inode_bitmap(ctx->inode_bb_map, ino);
+			ext2fs_mark_inode_bitmap(ctx->inode_used_map, ino);
+			continue;
+		}
+		if (pctx.errcode) {
+			fix_problem(ctx, PR_1_ISCAN_ERROR, &pctx);
+			ctx->flags |= E2F_FLAG_ABORT;
+			return;
+		}
+		if (!ino)
+			break;
 		pctx.ino = ino;
 		pctx.inode = &inode;
 		ctx->stashed_ino = ino;
@@ -401,7 +436,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 			pb.num_blocks = pb.last_block = 0;
 			pb.num_illegal_blocks = 0;
 			pb.suppress = 0; pb.clear = 0; pb.is_dir = 0;
-			pb.fragmented = 0;
+			pb.is_reg = 0; pb.fragmented = 0;
 			pb.inode = &inode;
 			pb.pctx = &pctx;
 			pb.ctx = ctx;
@@ -414,7 +449,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 			}
 			ext2fs_mark_inode_bitmap(ctx->inode_used_map, ino);
 			clear_problem_context(&pctx);
-			goto next;
+			continue;
 		} else if (ino == EXT2_ROOT_INO) {
 			/*
 			 * Make sure the root inode is a directory; if
@@ -451,11 +486,6 @@ void e2fsck_pass1(e2fsck_t ctx)
 		} else if (ino == EXT2_JOURNAL_INO) {
 			ext2fs_mark_inode_bitmap(ctx->inode_used_map, ino);
 			if (fs->super->s_journal_inum == EXT2_JOURNAL_INO) {
-				/*
-				 * XXX arguably this check should be
-				 * in journal.c, before we decide it's
-				 * safe to run the journal...
-				 */
 				if (!LINUX_S_ISREG(inode.i_mode) &&
 				    fix_problem(ctx, PR_1_JOURNAL_BAD_MODE,
 						&pctx)) {
@@ -464,7 +494,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 							   "pass1");
 				}
 				check_blocks(ctx, &pctx, block_buf);
-				goto next;
+				continue;
 			}
 			if ((inode.i_links_count || inode.i_blocks ||
 			     inode.i_blocks || inode.i_block[0]) &&
@@ -498,7 +528,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 				}
 			}
 			check_blocks(ctx, &pctx, block_buf);
-			goto next;
+			continue;
 		}
 		/*
 		 * Check for inodes who might have been part of the
@@ -517,7 +547,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 		 * than nothing.  The right answer is that there
 		 * shouldn't be any bugs in the orphan list handling.  :-)
 		 */
-		if (inode.i_dtime &&
+		if (inode.i_dtime && !busted_fs_time &&
 		    inode.i_dtime < ctx->fs->super->s_inodes_count) {
 			if (fix_problem(ctx, PR_1_LOW_DTIME, &pctx)) {
 				inode.i_dtime = inode.i_links_count ?
@@ -540,7 +570,7 @@ void e2fsck_pass1(e2fsck_t ctx)
 							   "pass1");
 				}
 			}
-			goto next;
+			continue;
 		}
 		/*
 		 * n.b.  0.3c ext2fs code didn't clear i_links_count for
@@ -603,32 +633,32 @@ void e2fsck_pass1(e2fsck_t ctx)
 			ext2fs_mark_inode_bitmap(ctx->inode_reg_map, ino);
 			ctx->fs_regular_count++;
 		} else if (LINUX_S_ISCHR (inode.i_mode) &&
-			   e2fsck_pass1_check_device_inode(&inode)) {
+			   e2fsck_pass1_check_device_inode(fs, &inode)) {
 			check_immutable(ctx, &pctx);
 			check_size(ctx, &pctx);
 			ctx->fs_chardev_count++;
 		} else if (LINUX_S_ISBLK (inode.i_mode) &&
-			   e2fsck_pass1_check_device_inode(&inode)) {
+			   e2fsck_pass1_check_device_inode(fs, &inode)) {
 			check_immutable(ctx, &pctx);
 			check_size(ctx, &pctx);
 			ctx->fs_blockdev_count++;
-		} else if (LINUX_S_ISLNK (inode.i_mode)) {
+		} else if (LINUX_S_ISLNK (inode.i_mode) &&
+			   e2fsck_pass1_check_symlink(fs, &inode, block_buf)) {
 			check_immutable(ctx, &pctx);
 			ctx->fs_symlinks_count++;
-			if (!e2fsck_pass1_check_symlink(fs, &inode))
-				mark_inode_bad(ctx, ino);
-			if (!inode.i_blocks) {
+			if (ext2fs_inode_data_blocks(fs, &inode) == 0) {
 				ctx->fs_fast_symlinks_count++;
-				goto next;
+				check_blocks(ctx, &pctx, block_buf);
+				continue;
 			}
 		}
 		else if (LINUX_S_ISFIFO (inode.i_mode) &&
-			 e2fsck_pass1_check_device_inode(&inode)) {
+			 e2fsck_pass1_check_device_inode(fs, &inode)) {
 			check_immutable(ctx, &pctx);
 			check_size(ctx, &pctx);
 			ctx->fs_fifo_count++;
 		} else if ((LINUX_S_ISSOCK (inode.i_mode)) &&
-			   e2fsck_pass1_check_device_inode(&inode)) {
+			   e2fsck_pass1_check_device_inode(fs, &inode)) {
 			check_immutable(ctx, &pctx);
 			check_size(ctx, &pctx);
 			ctx->fs_sockets_count++;
@@ -659,22 +689,6 @@ void e2fsck_pass1(e2fsck_t ctx)
 			if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
 				return;
 		}
-	next:
-		pctx.errcode = ext2fs_get_next_inode(scan, &ino, &inode);
-		if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
-			return;
-		if (pctx.errcode == EXT2_ET_BAD_BLOCK_IN_INODE_TABLE) {
-			if (!ctx->inode_bb_map)
-				alloc_bb_map(ctx);
-			ext2fs_mark_inode_bitmap(ctx->inode_bb_map, ino);
-			ext2fs_mark_inode_bitmap(ctx->inode_used_map, ino);
-			goto next;
-		}
-		if (pctx.errcode) {
-			fix_problem(ctx, PR_1_ISCAN_ERROR, &pctx);
-			ctx->flags |= E2F_FLAG_ABORT;
-			return;
-		}
 	}
 	process_inodes(ctx, block_buf);
 	ext2fs_close_inode_scan(scan);
@@ -1057,10 +1071,14 @@ static int check_ext_attr(e2fsck_t ctx, struct problem_context *pctx,
 		goto clear_extattr;
 	header = (struct ext2_ext_attr_header *) block_buf;
 	pctx->blk = inode->i_file_acl;
-	if (header->h_magic != EXT2_EXT_ATTR_MAGIC) {
+	if (((ctx->ext_attr_ver == 1) &&
+	     (header->h_magic != EXT2_EXT_ATTR_MAGIC_v1)) ||
+	    ((ctx->ext_attr_ver == 2) &&
+	     (header->h_magic != EXT2_EXT_ATTR_MAGIC))) {
 		if (fix_problem(ctx, PR_1_BAD_EA_BLOCK, pctx))
 			goto clear_extattr;
 	}
+
 	if (header->h_blocks != 1) {
 		if (fix_problem(ctx, PR_1_EA_MULTI_BLOCK, pctx))
 			goto clear_extattr;
@@ -1085,7 +1103,10 @@ static int check_ext_attr(e2fsck_t ctx, struct problem_context *pctx,
 			if (fix_problem(ctx, PR_1_EA_ALLOC_COLLISION, pctx))
 				goto clear_extattr;
 		}
-		if (entry->e_name_len == 0 || entry->e_name_index != 0) {
+		if ((ctx->ext_attr_ver == 1 &&
+		     (entry->e_name_len == 0 || entry->e_name_index != 0)) ||
+		    (ctx->ext_attr_ver == 2 &&
+		     entry->e_name_index == 0)) {
 			if (fix_problem(ctx, PR_1_EA_BAD_NAME, pctx))
 				goto clear_extattr;
 		}
@@ -1121,6 +1142,58 @@ clear_extattr:
 	return 0;
 }
 
+/* Returns 1 if bad htree, 0 if OK */
+static int handle_htree(e2fsck_t ctx, struct problem_context *pctx,
+			ext2_ino_t ino, struct ext2_inode *inode,
+			char *block_buf)
+{
+	struct ext2_dx_root_info	*root;
+	ext2_filsys			fs = ctx->fs;
+	errcode_t			retval;
+	blk_t				blk;
+
+	if ((!LINUX_S_ISDIR(inode->i_mode) &&
+	     fix_problem(ctx, PR_1_HTREE_NODIR, pctx)) ||
+	    (!(fs->super->s_feature_compat & EXT2_FEATURE_COMPAT_DIR_INDEX) &&
+	     fix_problem(ctx, PR_1_HTREE_SET, pctx)))
+		return 1;
+
+	blk = inode->i_block[0];
+	if (((blk == 0) ||
+	     (blk < fs->super->s_first_data_block) ||
+	     (blk >= fs->super->s_blocks_count)) &&
+	    fix_problem(ctx, PR_1_HTREE_BADROOT, pctx))
+		return 1;
+
+	retval = io_channel_read_blk(fs->io, blk, 1, block_buf);
+	if (retval && fix_problem(ctx, PR_1_HTREE_BADROOT, pctx))
+		return 1;
+	
+	/* XXX should check that beginning matches a directory */
+	root = (struct ext2_dx_root_info *) (block_buf + 24);
+
+	if ((root->reserved_zero || root->info_length < 8) &&
+	    fix_problem(ctx, PR_1_HTREE_BADROOT, pctx))
+		return 1;
+
+	pctx->num = root->hash_version;
+	if ((root->hash_version != EXT2_HASH_LEGACY) &&
+	    (root->hash_version != EXT2_HASH_HALF_MD4) &&
+	    (root->hash_version != EXT2_HASH_TEA) &&
+	    fix_problem(ctx, PR_1_HTREE_HASHV, pctx))
+		return 1;
+		
+	if ((root->unused_flags & EXT2_HASH_FLAG_INCOMPAT) &&
+	    fix_problem(ctx, PR_1_HTREE_INCOMPAT, pctx))
+		return 1;
+
+	pctx->num = root->indirect_levels;
+	if ((root->indirect_levels > 1) &&
+	    fix_problem(ctx, PR_1_HTREE_DEPTH, pctx))
+		return 1;
+	
+	return 0;
+}
 
 /*
  * This subroutine is called on each inode to account for all of the
@@ -1134,23 +1207,25 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 	ext2_ino_t	ino = pctx->ino;
 	struct ext2_inode *inode = pctx->inode;
 	int		bad_size = 0;
+	int		dirty_inode = 0;
 	__u64		size;
 	
-	if (!ext2fs_inode_has_valid_blocks(pctx->inode))
-		return;
-	
 	pb.ino = ino;
-	pb.num_blocks = pb.last_block = 0;
+	pb.num_blocks = 0;
+	pb.last_block = -1;
 	pb.num_illegal_blocks = 0;
 	pb.suppress = 0; pb.clear = 0;
 	pb.fragmented = 0;
 	pb.compressed = 0;
 	pb.previous_block = 0;
-	pb.is_dir = LINUX_S_ISDIR(pctx->inode->i_mode);
+	pb.is_dir = LINUX_S_ISDIR(inode->i_mode);
+	pb.is_reg = LINUX_S_ISREG(inode->i_mode);
+	pb.max_blocks = 1 << (31 - fs->super->s_log_block_size);
 	pb.inode = inode;
 	pb.pctx = pctx;
 	pb.ctx = ctx;
 	pctx->ino = ino;
+	pctx->errcode = 0;
 
 	if (inode->i_flags & EXT2_COMPRBLK_FL) {
 		if (fs->super->s_feature_incompat &
@@ -1159,18 +1234,19 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 		else {
 			if (fix_problem(ctx, PR_1_COMPR_SET, pctx)) {
 				inode->i_flags &= ~EXT2_COMPRBLK_FL;
-				e2fsck_write_inode(ctx, ino, inode,
-						   "check_blocks");
+				dirty_inode++;
 			}
 		}
 	}
 
-	pctx->errcode = ext2fs_block_iterate2(fs, ino,
+	if (ext2fs_inode_has_valid_blocks(inode))
+		pctx->errcode = ext2fs_block_iterate2(fs, ino,
 				       pb.is_dir ? BLOCK_FLAG_HOLE : 0,
 				       block_buf, process_block, &pb);
-	if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
-		return;
 	end_problem_latch(ctx, PR_LATCH_BLOCK);
+	end_problem_latch(ctx, PR_LATCH_TOOBIG);
+	if (ctx->flags & E2F_FLAG_SIGNAL_MASK)
+		goto out;
 	if (pctx->errcode)
 		fix_problem(ctx, PR_1_BLOCK_ITERATE, pctx);
 
@@ -1178,11 +1254,10 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 		ctx->fs_fragmented++;
 
 	if (pb.clear) {
-		e2fsck_read_inode(ctx, ino, inode, "check_blocks");
 		inode->i_links_count = 0;
 		ext2fs_icount_store(ctx->inode_link_info, ino, 0);
 		inode->i_dtime = time(0);
-		e2fsck_write_inode(ctx, ino, inode, "check_blocks");
+		dirty_inode++;
 		ext2fs_unmark_inode_bitmap(ctx->inode_dir_map, ino);
 		ext2fs_unmark_inode_bitmap(ctx->inode_reg_map, ino);
 		ext2fs_unmark_inode_bitmap(ctx->inode_used_map, ino);
@@ -1192,8 +1267,38 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 		 * restart the pass 1 scan.
 		 */
 		ctx->flags |= E2F_FLAG_RESTART;
-		return;
+		goto out;
 	}
+	
+	if (inode->i_flags & EXT2_INDEX_FL) {
+		if (handle_htree(ctx, pctx, ino, inode, block_buf)) {
+			inode->i_flags &= ~EXT2_INDEX_FL;
+			dirty_inode++;
+		} else {
+#ifdef ENABLE_HTREE
+			e2fsck_add_dx_dir(ctx, ino, pb.last_block+1);
+#endif
+		}
+	}
+	if (ctx->dirs_to_hash && pb.is_dir &&
+	    !(inode->i_flags & EXT2_INDEX_FL) &&
+	    ((inode->i_size / fs->blocksize) >= 3))
+		ext2fs_u32_list_add(ctx->dirs_to_hash, ino);
+		
+	if (!pb.num_blocks && pb.is_dir) {
+		if (fix_problem(ctx, PR_1_ZERO_LENGTH_DIR, pctx)) {
+			inode->i_links_count = 0;
+			ext2fs_icount_store(ctx->inode_link_info, ino, 0);
+			inode->i_dtime = time(0);
+			dirty_inode++;
+			ext2fs_unmark_inode_bitmap(ctx->inode_dir_map, ino);
+			ext2fs_unmark_inode_bitmap(ctx->inode_reg_map, ino);
+			ext2fs_unmark_inode_bitmap(ctx->inode_used_map, ino);
+			ctx->fs_directory_count--;
+			goto out;
+		}
+	}
+
 	if (inode->i_file_acl && check_ext_attr(ctx, pctx, block_buf))
 		pb.num_blocks++;
 
@@ -1203,24 +1308,9 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 	       ino, inode->i_size, pb.last_block, inode->i_blocks,
 	       pb.num_blocks);
 #endif
-	if (!pb.num_blocks && pb.is_dir) {
-		if (fix_problem(ctx, PR_1_ZERO_LENGTH_DIR, pctx)) {
-			inode->i_links_count = 0;
-			ext2fs_icount_store(ctx->inode_link_info, ino, 0);
-			inode->i_dtime = time(0);
-			e2fsck_write_inode(ctx, ino, inode, "check_blocks");
-			ext2fs_unmark_inode_bitmap(ctx->inode_dir_map, ino);
-			ext2fs_unmark_inode_bitmap(ctx->inode_reg_map, ino);
-			ext2fs_unmark_inode_bitmap(ctx->inode_used_map, ino);
-			ctx->fs_directory_count--;
-			pb.is_dir = 0;
-		}
-	}
 	if (pb.is_dir) {
 		int nblock = inode->i_size >> EXT2_BLOCK_SIZE_BITS(fs->super);
-		/* We don't let a directory become larger than 2GB */
-		if (nblock > (pb.last_block + 1) ||
-		    (inode->i_size & ((fs->blocksize-1) | 0x80000000UL)) != 0)
+		if (nblock > (pb.last_block + 1))
 			bad_size = 1;
 		else if (nblock < (pb.last_block + 1)) {
 			if (((pb.last_block + 1) - nblock) >
@@ -1229,32 +1319,37 @@ static void check_blocks(e2fsck_t ctx, struct problem_context *pctx,
 		}
 	} else {
 		size = inode->i_size | ((__u64) inode->i_size_high << 32);
-		if ((size < pb.last_block * fs->blocksize))
+		if ((pb.last_block >= 0) &&
+		    (size < pb.last_block * fs->blocksize))
 			bad_size = 3;
 		else if (size > ext2_max_sizes[fs->super->s_log_block_size])
 			bad_size = 4;
-		/* FIXME: need to ensure pb.num_blocks < 2^32 */
 	}
-	if (bad_size) {
+	/* i_size for symlinks is checked elsewhere */
+	if (bad_size && !LINUX_S_ISLNK(inode->i_mode)) {
 		pctx->num = (pb.last_block+1) * fs->blocksize;
 		if (fix_problem(ctx, PR_1_BAD_I_SIZE, pctx)) {
 			inode->i_size = pctx->num;
-			if (!pb.is_dir)
+			if (!LINUX_S_ISDIR(inode->i_mode))
 				inode->i_size_high = pctx->num >> 32;
-			e2fsck_write_inode(ctx, ino, inode, "check_blocks");
+			dirty_inode++;
 		}
 		pctx->num = 0;
 	}
-	if (!pb.is_dir && (inode->i_size_high || inode->i_size & 0x80000000UL))
+	if (LINUX_S_ISREG(inode->i_mode) &&
+	    (inode->i_size_high || inode->i_size & 0x80000000UL))
 		ctx->large_files++;
 	if (pb.num_blocks != inode->i_blocks) {
 		pctx->num = pb.num_blocks;
 		if (fix_problem(ctx, PR_1_BAD_I_BLOCKS, pctx)) {
 			inode->i_blocks = pb.num_blocks;
-			e2fsck_write_inode(ctx, ino, inode, "check_blocks");
+			dirty_inode++;
 		}
 		pctx->num = 0;
 	}
+out:
+	if (dirty_inode)
+		e2fsck_write_inode(ctx, ino, inode, "check_blocks");
 }
 
 #if 0
@@ -1351,7 +1446,7 @@ static int process_block(ext2_filsys fs,
 		   perhaps this really is just an illegal block. */
 		return 0;
 	}
-	
+
 	if (blk == 0) {
 		if (p->is_dir == 0) {
 			/*
@@ -1392,7 +1487,14 @@ static int process_block(ext2_filsys fs,
 			p->fragmented = 1;
 	}
 	p->previous_block = blk;
-	
+
+	if (p->is_dir && blockcnt > (1 << (15 - fs->super->s_log_block_size)))
+		problem = PR_1_TOOBIG_DIR;
+	if (p->is_reg && p->num_blocks+1 >= p->max_blocks)
+		problem = PR_1_TOOBIG_REG;
+	if (!p->is_dir && !p->is_reg && blockcnt > 0)
+		problem = PR_1_TOOBIG_SYMLINK;
+	    
 	if (blk < fs->super->s_first_data_block ||
 	    blk >= fs->super->s_blocks_count)
 		problem = PR_1_ILLEGAL_BLOCK_NUM;
@@ -1630,6 +1732,7 @@ static void new_table_block(e2fsck_t ctx, blk_t first_block, int group,
 		return;
 	}
 	ext2fs_mark_super_dirty(fs);
+	fs->flags &= ~EXT2_FLAG_MASTER_SB_ONLY;
 	pctx.blk2 = *new_block;
 	fix_problem(ctx, (old_block ? PR_1_RELOC_FROM_TO :
 			  PR_1_RELOC_TO), &pctx);
@@ -1695,30 +1798,52 @@ static void mark_table_blocks(e2fsck_t ctx)
 {
 	ext2_filsys fs = ctx->fs;
 	blk_t	block, b;
-	int	i,j;
+	int	i, j, has_super, meta_bg, meta_bg_size, old_desc_blocks;
 	struct problem_context pctx;
 	
 	clear_problem_context(&pctx);
 	
 	block = fs->super->s_first_data_block;
+	if (fs->super->s_feature_incompat & EXT2_FEATURE_INCOMPAT_META_BG)
+		old_desc_blocks = fs->super->s_first_meta_bg;
+	else
+		old_desc_blocks = fs->desc_blocks;
 	for (i = 0; i < fs->group_desc_count; i++) {
 		pctx.group = i;
 
-		if (ext2fs_bg_has_super(fs, i)) {
+		has_super = ext2fs_bg_has_super(fs, i);
+		if (has_super)
 			/*
 			 * Mark this group's copy of the superblock
 			 */
 			ext2fs_mark_block_bitmap(ctx->block_found_map, block);
 		
-			/*
-			 * Mark this group's copy of the descriptors
-			 */
-			for (j = 0; j < fs->desc_blocks; j++) {
-				ext2fs_mark_block_bitmap(ctx->block_found_map,
+		meta_bg_size = (fs->blocksize /
+				sizeof (struct ext2_group_desc));
+		meta_bg = i / meta_bg_size;
+		if (!(fs->super->s_feature_incompat &
+		      EXT2_FEATURE_INCOMPAT_META_BG) ||
+		    (meta_bg < fs->super->s_first_meta_bg)) {
+			if (has_super) {
+				/*
+				 * Mark this group's copy of the descriptors
+				 */
+				for (j = 0; j < old_desc_blocks; j++) {
+					ext2fs_mark_block_bitmap(ctx->block_found_map,
 							 block + j + 1);
+				}
 			}
+		} else {
+			if (has_super)
+				has_super = 1;
+			if (((i % meta_bg_size) == 0) ||
+			    ((i % meta_bg_size) == 1) ||
+			    ((i % meta_bg_size) == (meta_bg_size-1)))
+				ext2fs_mark_block_bitmap(ctx->block_found_map,
+						 block + has_super);
 		}
 		
+		
 		/*
 		 * Mark the blocks used for the inode table
 		 */
