@@ -36,6 +36,8 @@ struct problem_context {
 #define PR_LATCH_RELOC	0x0050  /* Latch for superblock relocate hint */
 #define PR_LATCH_DBLOCK	0x0060	/* Latch for pass 1b dup block headers */
 #define PR_LATCH_LOW_DTIME 0x0070 /* Latch for pass1 orphaned list refugees */
+#define PR_LATCH_TOOBIG	0x0080	/* Latch for file to big errors */
+#define PR_LATCH_OPTIMIZE_DIR 0x0090 /* Latch for optimize directories */
 
 #define PR_LATCH(x)	((((x) & PR_LATCH_MASK) >> 4) - 1)
 
@@ -362,16 +364,46 @@ struct problem_context {
 
 /* Error allocating EA region allocation structure */
 #define PR_1_EA_ALLOC_REGION		0x01003F
-	
+
 /* Error EA allocation collision */
 #define PR_1_EA_ALLOC_COLLISION		0x010040
-	
+
 /* Bad extended attribute name */
 #define PR_1_EA_BAD_NAME		0x010041
 
 /* Bad extended attribute value */
-#define PR_1_EA_BAD_VALUE		0x0100423
-	
+#define PR_1_EA_BAD_VALUE		0x010042
+
+/* Inode too big (latch question) */
+#define PR_1_INODE_TOOBIG		0x010043
+
+/* Directory too big */
+#define PR_1_TOOBIG_DIR			0x010044
+
+/* Regular file too big */
+#define PR_1_TOOBIG_REG			0x010045
+
+/* Symlink too big */
+#define PR_1_TOOBIG_SYMLINK		0x010046
+
+/* INDEX_FL flag set on a non-HTREE filesystem */
+#define PR_1_HTREE_SET			0x010047
+
+/* INDEX_FL flag set on a non-directory */	
+#define PR_1_HTREE_NODIR		0x010048
+
+/* Invalid root node in HTREE directory */	
+#define PR_1_HTREE_BADROOT		0x010049
+
+/* Unsupported hash version in HTREE directory */	
+#define PR_1_HTREE_HASHV		0x01004A
+
+/* Incompatible flag in HTREE root node */	
+#define PR_1_HTREE_INCOMPAT		0x01004B
+
+/* HTREE too deep */	
+#define PR_1_HTREE_DEPTH		0x01004C
+
 /*
  * Pass 1b errors
  */
@@ -397,7 +429,10 @@ struct problem_context {
 /* Error while iterating over blocks */
 #define PR_1B_BLOCK_ITERATE	0x0110006
 
-	
+/* Error adjusting EA refcount */
+#define PR_1B_ADJ_EA_REFCOUNT	0x0110007
+
+
 /* Pass 1C: Scan directories for inodes with dup blocks. */
 #define PR_1C_PASS_HEADER	0x012000
 
@@ -562,8 +597,8 @@ struct problem_context {
 /* Directory filename can't be zero-length  */
 #define PR_2_NULL_NAME		0x020030
 
-/* Invalid fast symlink size */
-#define PR_2_SYMLINK_SIZE	0x020031
+/* Invalid symlink */
+#define PR_2_INVALID_SYMLINK	0x020031
 
 /* i_file_acl (extended attribute) is bad */
 #define PR_2_FILE_ACL_BAD	0x020032
@@ -571,6 +606,45 @@ struct problem_context {
 /* Filesystem contains large files, but has no such flag in sb */
 #define PR_2_FEATURE_LARGE_FILES 0x020033
 
+/* Node in HTREE directory not referenced */
+#define PR_2_HTREE_NOTREF	0x020034
+
+/* Node in HTREE directory referenced twice */
+#define PR_2_HTREE_DUPREF	0x020035
+
+/* Node in HTREE directory has bad min hash */
+#define PR_2_HTREE_MIN_HASH	0x020036
+
+/* Node in HTREE directory has bad max hash */
+#define PR_2_HTREE_MAX_HASH	0x020037
+
+/* Clear invalid HTREE directory */
+#define PR_2_HTREE_CLEAR	0x020038
+
+/* Clear the htree flag forcibly */
+#define PR_2_HTREE_FCLR		0x020039
+
+/* Bad block in htree interior node */
+#define PR_2_HTREE_BADBLK	0x02003A
+
+/* Error adjusting EA refcount */
+#define PR_2_ADJ_EA_REFCOUNT	0x02003B
+
+/* Invalid HTREE root node */
+#define PR_2_HTREE_BAD_ROOT	0x02003C
+
+/* Invalid HTREE limit */
+#define PR_2_HTREE_BAD_LIMIT	0x02003D
+
+/* Invalid HTREE count */
+#define PR_2_HTREE_BAD_COUNT	0x02003E
+
+/* HTREE interior node has out-of-order hashes in table */
+#define PR_2_HTREE_HASH_ORDER	0x02003F
+
+/* Node in HTREE directory has bad depth */
+#define PR_2_HTREE_BAD_DEPTH	0x020040	
+
 /*
  * Pass 3 errors
  */
@@ -647,6 +721,27 @@ struct problem_context {
 /* Lost+found is not a directory */
 #define PR_3_LPF_NOTDIR			0x030017
 
+/*
+ * Pass 3a --- rehashing diretories
+ */
+/* Pass 3a: Reindexing directories */
+#define PR_3A_PASS_HEADER		0x031000
+
+/* Error iterating over directories */
+#define PR_3A_OPTIMIZE_ITER		0x031001
+
+/* Error rehash directory */
+#define PR_3A_OPTIMIZE_DIR_ERR		0x031002		
+
+/* Rehashing dir header */
+#define PR_3A_OPTIMIZE_DIR_HEADER		0x031003
+
+/* Rehashing directory %d */
+#define PR_3A_OPTIMIZE_DIR		0x031004
+		  
+/* Rehashing dir end */	  
+#define PR_3A_OPTIMIZE_DIR_END		0x031005
+
 /*
  * Pass 4 errors
  */
