@@ -29,11 +29,7 @@
 #include <sys/types.h>
 #include <sys/time.h>
 
-#ifdef HAVE_LINUX_FS_H
-#include <linux/fs.h>
-#endif
-#include <linux/ext2_fs.h>
-
+#include "ext2fs/ext2_fs.h"
 #include "ext2fs/ext2fs.h"
 
 
@@ -80,7 +76,7 @@ static void print_resource_track(struct resource_track *track)
 	gettimeofday(&time_end, 0);
 	getrusage(RUSAGE_SELF, &r);
 
-	printf("Memory used: %d, elapsed time: %6.3f/%6.3f/%6.3f\n",
+	printf(_("Memory used: %d, elapsed time: %6.3f/%6.3f/%6.3f\n"),
 	       (int) (((char *) sbrk(0)) - ((char *) track->brk_start)),
 	       timeval_subtract(&time_end, &track->time_start),
 	       timeval_subtract(&r.ru_utime, &track->user_start),
@@ -96,10 +92,10 @@ int main (int argc, char *argv[])
 	int		i;
 	ext2_filsys	fs;
 	ext2_inode_scan	scan;
-	ino_t	ino;
+	ext2_ino_t	ino;
 	struct ext2_inode inode;
 
-	printf("size of inode=%d\n", sizeof(inode));
+	printf(_("size of inode=%d\n"), sizeof(inode));
 
 	device_name = "/dev/hda3";
 	
@@ -108,19 +104,19 @@ int main (int argc, char *argv[])
 	retval = ext2fs_open(device_name, 0,
 			     0, 0, unix_io_manager, &fs);
 	if (retval) {
-		com_err(argv[0], retval, "while trying to open %s",
+		com_err(argv[0], retval, _("while trying to open %s"),
 			device_name);
 		exit(1);
 	}
 
 	retval = ext2fs_open_inode_scan(fs, 0, &scan);
 	if (retval) {
-		com_err(argv[0], retval, "while opening inode scan");
+		com_err(argv[0], retval, _("while opening inode scan"));
 		exit(1);
 	}
 	retval = ext2fs_get_next_inode(scan, &ino, &inode);
 	if (retval) {
-		com_err(argv[0], retval, "while starting inode scan");
+		com_err(argv[0], retval, _("while starting inode scan"));
 		exit(1);
 	}
 	while (ino) {
@@ -131,7 +127,7 @@ int main (int argc, char *argv[])
 		retval = ext2fs_get_next_inode(scan, &ino, &inode);
 		if (retval) {
 			com_err(argv[0], retval,
-				"while doing inode scan");
+				_("while doing inode scan"));
 			exit(1);
 		}
 	}
