@@ -152,10 +152,7 @@ static int release_inode_block(ext2_filsys fs,
 		retval |= BLOCK_CHANGED;
 	}
 	
-	ext2fs_unmark_block_bitmap(fs->block_map, blk);
-	fs->group_desc[ext2fs_group_of_blk(fs, blk)].bg_free_blocks_count++;
-	fs->super->s_free_blocks_count++;
-	
+	ext2fs_block_alloc_stats(fs, blk, -1);
 	return retval;
 }
 		
@@ -168,9 +165,10 @@ static int release_inode_blocks(e2fsck_t ctx, ext2_ino_t ino,
 				struct ext2_inode *inode, char *block_buf,
 				struct problem_context *pctx)
 {
+	struct process_block_struct 	pb;
 	ext2_filsys			fs = ctx->fs;
 	errcode_t			retval;
-	struct process_block_struct 	pb;
+	__u32				count;
 
 	if (!ext2fs_inode_has_valid_blocks(inode))
 		return 0;
@@ -211,7 +209,23 @@ static int release_inode_blocks(e2fsck_t ctx, ext2_ino_t ino,
 		inode->i_blocks -= pb.truncated_blocks *
 			(fs->blocksize / 512);
 
-	ext2fs_mark_bb_dirty(fs);
+	if (inode->i_file_acl) {
+		retval = ext2fs_adjust_ea_refcount(fs, inode->i_file_acl,
+						   block_buf, -1, &count);
+		if (retval == EXT2_ET_BAD_EA_BLOCK_NUM) {
+			retval = 0;
+			count = 1;
+		}
+		if (retval) {
+			com_err("release_inode_blocks", retval,
+		_("while calling ext2fs_adjust_ea_refocunt for inode %d"),
+				ino);
+			return 1;
+		}
+		if (count == 0)
+			ext2fs_block_alloc_stats(fs, inode->i_file_acl, -1);
+		inode->i_file_acl = 0;
+	}
 	return 0;
 }
 
@@ -222,7 +236,6 @@ static int release_inode_blocks(e2fsck_t ctx, ext2_ino_t ino,
 static int release_orphan_inodes(e2fsck_t ctx)
 {
 	ext2_filsys fs = ctx->fs;
-	int group;
 	ext2_ino_t	ino, next_ino;
 	struct ext2_inode inode;
 	struct problem_context pctx;
@@ -237,6 +250,14 @@ static int release_orphan_inodes(e2fsck_t ctx)
 	 */
 	fs->super->s_last_orphan = 0;
 	ext2fs_mark_super_dirty(fs);
+
+	/*
+	 * If the filesystem contains errors, don't run the orphan
+	 * list, since the orphan list can't be trusted; and we're
+	 * going to be running a full e2fsck run anyway...
+	 */
+	if (fs->super->s_state & EXT2_ERROR_FS)
+		return 0;
 	
 	if ((ino < EXT2_FIRST_INODE(fs->super)) ||
 	    (ino > fs->super->s_inodes_count)) {
@@ -273,19 +294,16 @@ static int release_orphan_inodes(e2fsck_t ctx)
 			goto return_abort;
 
 		if (!inode.i_links_count) {
-			ext2fs_unmark_inode_bitmap(fs->inode_map, ino);
-			ext2fs_mark_ib_dirty(fs);
-			group = ext2fs_group_of_ino(fs, ino);
-			fs->group_desc[group].bg_free_inodes_count++;
-			fs->super->s_free_inodes_count++;
-			if (LINUX_S_ISDIR(inode.i_mode))
-				fs->group_desc[group].bg_used_dirs_count--;
-			
+			ext2fs_inode_alloc_stats2(fs, ino, -1,
+						  LINUX_S_ISDIR(inode.i_mode));
 			inode.i_dtime = time(0);
+		} else {
+			inode.i_dtime = 0;
 		}
 		e2fsck_write_inode(ctx, ino, &inode, "delete_file");
 		ino = next_ino;
 	}
+	ext2fs_free_mem((void **) &block_buf);
 	return 0;
 return_abort:
 	ext2fs_free_mem((void **) &block_buf);
@@ -299,13 +317,20 @@ void check_super_block(e2fsck_t ctx)
 	blk_t	first_block, last_block;
 	struct ext2_super_block *sb = fs->super;
 	blk_t	blocks_per_group = fs->super->s_blocks_per_group;
+	blk_t	bpg_max;
 	int	inodes_per_block;
+	int	ipg_max;
 	dgrp_t	i;
 	blk_t	should_be;
 	struct problem_context	pctx;
-	
-	inodes_per_block = (EXT2_BLOCK_SIZE(fs->super) /
-			    EXT2_INODE_SIZE(fs->super));
+
+	inodes_per_block = EXT2_INODES_PER_BLOCK(fs->super);
+	ipg_max = inodes_per_block * (blocks_per_group - 4);
+	if (ipg_max > EXT2_MAX_INODES_PER_GROUP(sb))
+		ipg_max = EXT2_MAX_INODES_PER_GROUP(sb);
+	bpg_max = 8 * EXT2_BLOCK_SIZE(sb);
+	if (bpg_max > EXT2_MAX_BLOCKS_PER_GROUP(sb))
+		bpg_max = EXT2_MAX_BLOCKS_PER_GROUP(sb);
 
 	ctx->invalid_inode_bitmap_flag = (int *) e2fsck_allocate_memory(ctx,
 		 sizeof(int) * fs->group_desc_count, "invalid_inode_bitmap");
@@ -313,7 +338,7 @@ void check_super_block(e2fsck_t ctx)
 		 sizeof(int) * fs->group_desc_count, "invalid_block_bitmap");
 	ctx->invalid_inode_table_flag = (int *) e2fsck_allocate_memory(ctx,
 		sizeof(int) * fs->group_desc_count, "invalid_inode_table");
-		
+
 	clear_problem_context(&pctx);
 
 	/*
@@ -325,19 +350,18 @@ void check_super_block(e2fsck_t ctx)
 			  MIN_CHECK, 1, 0);
 	check_super_value(ctx, "first_data_block", sb->s_first_data_block,
 			  MAX_CHECK, 0, sb->s_blocks_count);
-	check_super_value(ctx, "log_frag_size", sb->s_log_frag_size,
-			  MAX_CHECK, 0, 2);
 	check_super_value(ctx, "log_block_size", sb->s_log_block_size,
-			  MIN_CHECK | MAX_CHECK, sb->s_log_frag_size,
-			  2);
+			  MIN_CHECK | MAX_CHECK, 0,
+			  EXT2_MAX_BLOCK_LOG_SIZE - EXT2_MIN_BLOCK_LOG_SIZE);
+	check_super_value(ctx, "log_frag_size", sb->s_log_frag_size,
+			  MIN_CHECK | MAX_CHECK, 0, sb->s_log_block_size);
 	check_super_value(ctx, "frags_per_group", sb->s_frags_per_group,
 			  MIN_CHECK | MAX_CHECK, sb->s_blocks_per_group,
-			  8 * EXT2_BLOCK_SIZE(sb));
+			  bpg_max);
 	check_super_value(ctx, "blocks_per_group", sb->s_blocks_per_group,
-			  MIN_CHECK | MAX_CHECK, 8, 8 * EXT2_BLOCK_SIZE(sb));
+			  MIN_CHECK | MAX_CHECK, 8, bpg_max);
 	check_super_value(ctx, "inodes_per_group", sb->s_inodes_per_group,
-			  MIN_CHECK | MAX_CHECK, inodes_per_block,
-			  inodes_per_block * (blocks_per_group-4));
+			  MIN_CHECK | MAX_CHECK, inodes_per_block, ipg_max);
 	check_super_value(ctx, "r_blocks_count", sb->s_r_blocks_count,
 			  MAX_CHECK, 0, sb->s_blocks_count / 4);
 
@@ -410,30 +434,33 @@ void check_super_block(e2fsck_t ctx)
 		if ((fs->group_desc[i].bg_block_bitmap < first_block) ||
 		    (fs->group_desc[i].bg_block_bitmap >= last_block)) {
 			pctx.blk = fs->group_desc[i].bg_block_bitmap;
-			if (fix_problem(ctx, PR_0_BB_NOT_GROUP, &pctx)) {
+			if (fix_problem(ctx, PR_0_BB_NOT_GROUP, &pctx))
 				fs->group_desc[i].bg_block_bitmap = 0;
-				ctx->invalid_block_bitmap_flag[i]++;
-				ctx->invalid_bitmaps++;
-			}
+		}
+		if (fs->group_desc[i].bg_block_bitmap == 0) {
+			ctx->invalid_block_bitmap_flag[i]++;
+			ctx->invalid_bitmaps++;
 		}
 		if ((fs->group_desc[i].bg_inode_bitmap < first_block) ||
 		    (fs->group_desc[i].bg_inode_bitmap >= last_block)) {
 			pctx.blk = fs->group_desc[i].bg_inode_bitmap;
-			if (fix_problem(ctx, PR_0_IB_NOT_GROUP, &pctx)) {
+			if (fix_problem(ctx, PR_0_IB_NOT_GROUP, &pctx))
 				fs->group_desc[i].bg_inode_bitmap = 0;
-				ctx->invalid_inode_bitmap_flag[i]++;
-				ctx->invalid_bitmaps++;
-			}
+		}
+		if (fs->group_desc[i].bg_inode_bitmap == 0) {
+			ctx->invalid_inode_bitmap_flag[i]++;
+			ctx->invalid_bitmaps++;
 		}
 		if ((fs->group_desc[i].bg_inode_table < first_block) ||
 		    ((fs->group_desc[i].bg_inode_table +
 		      fs->inode_blocks_per_group - 1) >= last_block)) {
 			pctx.blk = fs->group_desc[i].bg_inode_table;
-			if (fix_problem(ctx, PR_0_ITABLE_NOT_GROUP, &pctx)) {
+			if (fix_problem(ctx, PR_0_ITABLE_NOT_GROUP, &pctx))
 				fs->group_desc[i].bg_inode_table = 0;
-				ctx->invalid_inode_table_flag[i]++;
-				ctx->invalid_bitmaps++;
-			}
+		}
+		if (fs->group_desc[i].bg_inode_table == 0) {
+			ctx->invalid_inode_table_flag[i]++;
+			ctx->invalid_bitmaps++;
 		}
 		first_block += fs->super->s_blocks_per_group;
 		last_block += fs->super->s_blocks_per_group;
@@ -457,6 +484,7 @@ void check_super_block(e2fsck_t ctx)
 		if (fix_problem(ctx, PR_0_ADD_UUID, &pctx)) {
 			uuid_generate(sb->s_uuid);
 			ext2fs_mark_super_dirty(fs);
+			fs->flags &= ~EXT2_FLAG_MASTER_SB_ONLY;
 		}
 	}
 #endif
@@ -501,6 +529,10 @@ void check_super_block(e2fsck_t ctx)
 		ext2fs_mark_super_dirty(fs);
 	}
 
+	/*
+	 * Move the ext3 journal file, if necessary.
+	 */
+	e2fsck_move_ext3_journal(ctx);
 	return;
 }
 
