@@ -51,19 +51,19 @@ static int verbose = 0;
 static int replace_bad_blocks = 0;
 static char *bad_blocks_file = 0;
 
-static int possible_block_sizes[] = { 1024, 2048, 4096, 8192, 0};
-
-static int root_filesystem = 0;
-static int read_only_root = 0;
-
 e2fsck_t e2fsck_global_ctx;	/* Try your very best not to use this! */
 
+#ifdef CONFIG_JBD_DEBUG		/* Enabled by configure --enable-jfs-debug */
+int journal_enable_debug = -1;
+#endif
+
 static void usage(e2fsck_t ctx)
 {
 	fprintf(stderr,
-		_("Usage: %s [-panyrcdfvstFSV] [-b superblock] [-B blocksize]\n"
+		_("Usage: %s [-panyrcdfvstDFSV] [-b superblock] [-B blocksize]\n"
 		"\t\t[-I inode_buffer_blocks] [-P process_inode_size]\n"
-		"\t\t[-l|-L bad_blocks_file] [-C fd] [-j ext-journal] device\n"),
+		"\t\t[-l|-L bad_blocks_file] [-C fd] [-j ext-journal]\n"
+		"\t\t[-E extended-options] device\n"),
 		ctx->program_name);
 
 	fprintf(stderr, _("\nEmergency help:\n"
@@ -122,7 +122,7 @@ static void show_stats(e2fsck_t	ctx)
 	 */
 #ifdef ENABLE_NLS
 	printf (_("\n%8d inodes used (%d%%)\n"), inodes_used,
-		(inodes_used != 1), 100 * inodes_used / inodes);
+		100 * inodes_used / inodes);
 	printf (_("%8d non-contiguous inodes (%0d.%d%%)\n"),
 		ctx->fs_fragmented, frag_percent / 10, frag_percent % 10);
 	printf (_("         # of inodes with ind/dind/tind blocks: %d/%d/%d\n"),
@@ -202,9 +202,10 @@ static void show_stats(e2fsck_t	ctx)
 static void check_mount(e2fsck_t ctx)
 {
 	errcode_t	retval;
-	int		mount_flags, cont;
+	int		cont;
 
-	retval = ext2fs_check_if_mounted(ctx->filesystem_name, &mount_flags);
+	retval = ext2fs_check_if_mounted(ctx->filesystem_name,
+					 &ctx->mount_flags);
 	if (retval) {
 		com_err("ext2fs_check_if_mount", retval,
 			_("while determining whether %s is mounted."),
@@ -216,9 +217,9 @@ static void check_mount(e2fsck_t ctx)
 	 * If the filesystem isn't mounted, or it's the root filesystem
 	 * and it's mounted read-only, then everything's fine.
 	 */
-	if ((!(mount_flags & EXT2_MF_MOUNTED)) ||
-	    ((mount_flags & EXT2_MF_ISROOT) &&
-	     (mount_flags & EXT2_MF_READONLY)))
+	if ((!(ctx->mount_flags & EXT2_MF_MOUNTED)) ||
+	    ((ctx->mount_flags & EXT2_MF_ISROOT) &&
+	     (ctx->mount_flags & EXT2_MF_READONLY)))
 		return;
 
 	if (ctx->options & E2F_OPT_READONLY) {
@@ -438,8 +439,72 @@ static void signal_progress_off(int sig)
 	e2fsck_clear_progbar(ctx);
 	ctx->progress = 0;
 }
+
+static void signal_cancel(int sig)
+{
+	e2fsck_t ctx = e2fsck_global_ctx;
+
+	if (!ctx)
+		exit(FSCK_CANCELED);
+
+	ctx->flags |= E2F_FLAG_CANCEL;
+}
 #endif
 
+static void parse_extended_opts(e2fsck_t ctx, const char *opts)
+{
+	char	*buf, *token, *next, *p, *arg;
+	int	len, ea_ver;
+	int	extended_usage = 0;
+
+	len = strlen(opts);
+	buf = malloc(len+1);
+	if (!buf) {
+		fprintf(stderr, _("Couldn't allocate memory to parse "
+			"extended options!\n"));
+		exit(1);
+	}
+	strcpy(buf, opts);
+	for (token = buf; token && *token; token = next) {
+		p = strchr(token, ',');
+		next = 0;
+		if (p) {
+			*p = 0;
+			next = p+1;
+		} 
+		arg = strchr(token, '=');
+		if (arg) {
+			*arg = 0;
+			arg++;
+		}
+		if (strcmp(token, "ea_ver") == 0) {
+			if (!arg) {
+				extended_usage++;
+				continue;
+			}
+			ea_ver = strtoul(arg, &p, 0);
+			if (*p ||
+			    ((ea_ver != 1) && (ea_ver != 2))) {
+				fprintf(stderr,
+					_("Invalid EA version.\n"));
+				extended_usage++;
+				continue;
+			}
+			ctx->ext_attr_ver = ea_ver;
+		} else
+			extended_usage++;
+	}
+	if (extended_usage) {
+		fprintf(stderr, _("Extended options are separated by commas, "
+			"and may take an argument which\n"
+			"is set off by an equals ('=') sign.  "
+			"Valid raid options are:\n"
+			"\tea_ver=<ea_version (1 or 2)\n\n"));
+		exit(1);
+	}
+}	
+
+
 static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 {
 	int		flush = 0;
@@ -452,6 +517,7 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 #ifdef HAVE_SIGNAL_H
 	struct sigaction	sa;
 #endif
+	char		*extended_opts = 0;
 
 	retval = e2fsck_allocate_context(&ctx);
 	if (retval)
@@ -467,7 +533,7 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 		ctx->program_name = *argv;
 	else
 		ctx->program_name = "e2fsck";
-	while ((c = getopt (argc, argv, "panyrcC:B:dfvtFVM:b:I:j:P:l:L:N:Ss")) != EOF)
+	while ((c = getopt (argc, argv, "panyrcC:B:dE:fvtFVM:b:I:j:P:l:L:N:SsD")) != EOF)
 		switch (c) {
 		case 'C':
 			ctx->progress = e2fsck_update_progress;
@@ -486,6 +552,12 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 			} else
 				close(fd);
 			break;
+		case 'D':
+			ctx->options |= E2F_OPT_COMPRESS_DIRS;
+			break;
+		case 'E':
+			extended_opts = optarg;
+			break;
 		case 'p':
 		case 'a':
 			ctx->options |= E2F_OPT_PREEN;
@@ -588,9 +660,12 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 	if (optind != argc - 1)
 		usage(ctx);
 	if ((ctx->options & E2F_OPT_NO) && !bad_blocks_file &&
-	    !cflag && !swapfs)
+	    !cflag && !swapfs && !(ctx->options & E2F_OPT_COMPRESS_DIRS))
 		ctx->options |= E2F_OPT_READONLY;
 	ctx->filesystem_name = argv[optind];
+	if (extended_opts)
+		parse_extended_opts(ctx, extended_opts);
+	
 	if (flush) {
 		fd = open(ctx->filesystem_name, O_RDONLY, 0);
 		if (fd < 0) {
@@ -626,6 +701,9 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 	 * Set up signal action
 	 */
 	memset(&sa, 0, sizeof(struct sigaction));
+	sa.sa_handler = signal_cancel;
+	sigaction(SIGINT, &sa, 0);
+	sigaction(SIGTERM, &sa, 0);
 #ifdef SA_RESTART
 	sa.sa_flags = SA_RESTART;
 #endif
@@ -653,6 +731,10 @@ static errcode_t PRS(int argc, char *argv[], e2fsck_t *ret_ctx)
 		} else
 			putenv (PATH_SET);
 	}
+#ifdef CONFIG_JBD_DEBUG
+	if (getenv("E2FSCK_JBD_DEBUG"))
+		journal_enable_debug = atoi(getenv("E2FSCK_JBD_DEBUG"));
+#endif
 	return 0;
 }
 
@@ -663,7 +745,6 @@ int main (int argc, char *argv[])
 {
 	errcode_t	retval = 0;
 	int		exit_value = FSCK_OK;
-	int		i;
 	ext2_filsys	fs = 0;
 	io_manager	io_ptr;
 	struct ext2_super_block *sb;
@@ -742,10 +823,11 @@ restart:
 				     ctx->superblock, ctx->blocksize,
 				     io_ptr, &fs);
 	} else if (ctx->superblock) {
-		for (i=0; possible_block_sizes[i]; i++) {
+		int blocksize;
+		for (blocksize = EXT2_MIN_BLOCK_SIZE;
+		     blocksize <= EXT2_MAX_BLOCK_SIZE; blocksize *= 2) {
 			retval = ext2fs_open(ctx->filesystem_name, flags,
-					     ctx->superblock,
-					     possible_block_sizes[i],
+					     ctx->superblock, blocksize,
 					     io_ptr, &fs);
 			if (!retval)
 				break;
@@ -878,7 +960,16 @@ restart:
 		com_err(ctx->program_name, 0,
 			_("Warning: compression support is experimental.\n"));
 #endif
-	
+#ifndef ENABLE_HTREE
+	if (sb->s_feature_compat & EXT2_FEATURE_COMPAT_DIR_INDEX) {
+		com_err(ctx->program_name, 0,
+			_("E2fsck not compiled with HTREE support,\n\t"
+			  "but filesystem %s has HTREE directories.\n"),
+			ctx->device_name);
+		goto get_newer;
+	}
+#endif
+
 	/*
 	 * If the user specified a specific superblock, presumably the
 	 * master superblock has been trashed.  So we mark the
@@ -889,8 +980,11 @@ restart:
 		ext2fs_mark_super_dirty(fs);
 
 	/*
-	 * Don't overwrite the backup superblock and block
-	 * descriptors, until we're sure the filesystem is OK....
+	 * We only update the master superblock because (a) paranoia;
+	 * we don't want to corrupt the backup superblocks, and (b) we
+	 * don't need to update the mount count and last checked
+	 * fields in the backup superblock (the kernel doesn't
+	 * update the backup superblocks anyway).
 	 */
 	fs->flags |= EXT2_FLAG_MASTER_SB_ONLY;
 
@@ -951,44 +1045,50 @@ restart:
 		ext2fs_close(fs);
 		goto restart;
 	}
-	if (run_result & E2F_FLAG_SIGNAL_MASK)
-		fatal_error(ctx, 0);
-	if (run_result & E2F_FLAG_CANCEL)
-		ext2fs_unmark_valid(fs);
+	if (run_result & E2F_FLAG_CANCEL) {
+		printf(_("%s: e2fsck canceled.\n"), ctx->device_name ?
+		       ctx->device_name : ctx->filesystem_name);
+		exit_value |= FSCK_CANCELED;
+	}
+	if (run_result & E2F_FLAG_ABORT)
+		fatal_error(ctx, _("aborted"));
 
 #ifdef MTRACE
 	mtrace_print("Cleanup");
 #endif
 	if (ext2fs_test_changed(fs)) {
-		exit_value = FSCK_NONDESTRUCT;
+		exit_value |= FSCK_NONDESTRUCT;
 		if (!(ctx->options & E2F_OPT_PREEN))
 		    printf(_("\n%s: ***** FILE SYSTEM WAS MODIFIED *****\n"),
 			       ctx->device_name);
-		if (root_filesystem && !read_only_root) {
+		if (ctx->mount_flags & EXT2_MF_ISROOT) {
 			printf(_("%s: ***** REBOOT LINUX *****\n"),
 			       ctx->device_name);
-			exit_value = FSCK_REBOOT;
+			exit_value |= FSCK_REBOOT;
 		}
 	}
-	if (ext2fs_test_valid(fs))
-		fs->flags &= ~EXT2_FLAG_MASTER_SB_ONLY;
-	else {
+	if (!ext2fs_test_valid(fs)) {
 		printf(_("\n%s: ********** WARNING: Filesystem still has "
 			 "errors **********\n\n"), ctx->device_name);
-		exit_value = FSCK_UNCORRECTED;
+		exit_value |= FSCK_UNCORRECTED;
+		exit_value &= ~FSCK_NONDESTRUCT;
 	}
-	if (!(ctx->options & E2F_OPT_READONLY)) {
-		if (ext2fs_test_valid(fs)) {
-			if (!(sb->s_state & EXT2_VALID_FS))
-				exit_value = FSCK_NONDESTRUCT;
-			sb->s_state = EXT2_VALID_FS;
-		} else
-			sb->s_state &= ~EXT2_VALID_FS;
-		sb->s_mnt_count = 0;
-		sb->s_lastcheck = time(NULL);
-		ext2fs_mark_super_dirty(fs);
+	if (exit_value & FSCK_CANCELED)
+		exit_value &= ~FSCK_NONDESTRUCT;
+	else {
+		show_stats(ctx);
+		if (!(ctx->options & E2F_OPT_READONLY)) {
+			if (ext2fs_test_valid(fs)) {
+				if (!(sb->s_state & EXT2_VALID_FS))
+					exit_value |= FSCK_NONDESTRUCT;
+				sb->s_state = EXT2_VALID_FS;
+			} else
+				sb->s_state &= ~EXT2_VALID_FS;
+			sb->s_mnt_count = 0;
+			sb->s_lastcheck = time(NULL);
+			ext2fs_mark_super_dirty(fs);
+		}
 	}
-	show_stats(ctx);
 
 	e2fsck_write_bitmaps(ctx);
 	
