@@ -37,14 +37,18 @@
 int fgetversion (const char * name, unsigned long * version)
 {
 #if HAVE_EXT2_IOCTLS
-	int fd, r, ver;
+	int fd, r, ver, save_errno = 0;
 
 	fd = open (name, OPEN_FLAGS);
 	if (fd == -1)
-		return - 1;
+		return -1;
 	r = ioctl (fd, EXT2_IOC_GETVERSION, &ver);
+	if (r == -1)
+		save_errno = errno;
 	*version = ver;
 	close (fd);
+	if (save_errno)
+		errno = save_errno;
 	return r;
 #else /* ! HAVE_EXT2_IOCTLS */
 	extern int errno;
