@@ -23,9 +23,9 @@
 #if HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#if HAVE_CHFLAGS
-#include <sys/stat.h>		/* For the flag values.  */
-#else
+#include <sys/types.h>
+#include <sys/stat.h>
+#if HAVE_EXT2_IOCTLS
 #include <fcntl.h>
 #include <sys/ioctl.h>
 #endif
@@ -40,6 +40,7 @@
 
 int fsetflags (const char * name, unsigned long flags)
 {
+	struct stat buf;
 #if HAVE_CHFLAGS
 	unsigned long bsd_flags = 0;
 
@@ -59,19 +60,27 @@ int fsetflags (const char * name, unsigned long flags)
 	return chflags (name, bsd_flags);
 #else
 #if HAVE_EXT2_IOCTLS
-	int fd, r, f;
+	int fd, r, f, save_errno = 0;
 
+	if (!stat(name, &buf) &&
+	    !S_ISREG(buf.st_mode) && !S_ISDIR(buf.st_mode)) {
+		close(fd);
+		goto notsupp;
+	}
 	fd = open (name, OPEN_FLAGS);
 	if (fd == -1)
 		return -1;
 	f = (int) flags;
 	r = ioctl (fd, EXT2_IOC_SETFLAGS, &f);
+	if (r == -1)
+		save_errno = errno;
 	close (fd);
+	if (save_errno)
+		errno = save_errno;
 	return r;
-#else /* ! HAVE_EXT2_IOCTLS */
-	extern int errno;
+#endif /* HAVE_EXT2_IOCTLS */
+#endif
+notsupp:
 	errno = EOPNOTSUPP;
 	return -1;
-#endif /* ! HAVE_EXT2_IOCTLS */
-#endif
 }
