@@ -112,6 +112,30 @@ static void print_features(struct ext2_super_block * s, FILE *f)
 #endif
 }
 
+static void print_mntopts(struct ext2_super_block * s, FILE *f)
+{
+#ifdef EXT2_DYNAMIC_REV
+	int	i, printed=0;
+	__u32	mask = s->s_default_mount_opts, m;
+
+	fprintf(f, "Default mount options:   ");
+	if (mask & EXT3_DEFM_JMODE) {
+		fprintf(f, " %s", e2p_mntopt2string(mask & EXT3_DEFM_JMODE));
+		printed++;
+	}
+	for (i=0,m=1; i < 32; i++, m<<=1) {
+		if (m & EXT3_DEFM_JMODE)
+			continue;
+		if (mask & m) {
+			fprintf(f, " %s", e2p_mntopt2string(m));
+			printed++;
+		}
+	}
+	if (printed == 0)
+		fprintf(f, " (none)");
+	fprintf(f, "\n");
+#endif
+}
 
 
 #ifndef EXT2_INODE_SIZE
@@ -145,11 +169,7 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 	} else
 		strcpy(buf, "<not available>");
 	fprintf(f, "Last mounted on:          %s\n", buf);
-	if (!e2p_is_null_uuid(sb->s_uuid)) {
-		e2p_uuid_to_str(sb->s_uuid, buf);
-	} else
-		strcpy(buf, "<none>");
-	fprintf(f, "Filesystem UUID:          %s\n", buf);
+	fprintf(f, "Filesystem UUID:          %s\n", e2p_uuid2str(sb->s_uuid));
 	fprintf(f, "Filesystem magic number:  0x%04X\n", sb->s_magic);
 	fprintf(f, "Filesystem revision #:    %d", sb->s_rev_level);
 	if (sb->s_rev_level == EXT2_GOOD_OLD_REV) {
@@ -161,6 +181,7 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 	} else
 		fprintf(f, " (unknown)\n");
 	print_features(sb, f);
+	print_mntopts(sb, f);
 	fprintf(f, "Filesystem state:        ");
 	print_fs_state (f, sb->s_state);
 	fprintf(f, "\n");
@@ -186,8 +207,16 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 	fprintf(f, "Fragments per group:      %u\n", sb->s_frags_per_group);
 	fprintf(f, "Inodes per group:         %u\n", sb->s_inodes_per_group);
 	fprintf(f, "Inode blocks per group:   %u\n", inode_blocks_per_group);
+	if (sb->s_first_meta_bg)
+		fprintf(f, "First meta block group:   %u\n",
+			sb->s_first_meta_bg);
+	if (sb->s_mkfs_time) {
+		tm = sb->s_mkfs_time;
+		fprintf(f, "Filesystem created:       %s", ctime(&tm));
+	}
 	tm = sb->s_mtime;
-	fprintf(f, "Last mount time:          %s", ctime(&tm));
+	fprintf(f, "Last mount time:          %s",
+		sb->s_mtime ? ctime(&tm) : "n/a\n");
 	tm = sb->s_wtime;
 	fprintf(f, "Last write time:          %s", ctime(&tm));
 	fprintf(f, "Mount count:              %u\n", sb->s_mnt_count);
@@ -212,15 +241,18 @@ void list_super2(struct ext2_super_block * sb, FILE *f)
 		fprintf(f, "Inode size:		  %d\n", sb->s_inode_size);
 	}
 	if (sb->s_feature_compat & EXT3_FEATURE_COMPAT_HAS_JOURNAL) {
-		if (e2p_is_null_uuid(sb->s_journal_uuid)) {
-			strcpy(buf, "<none>");
-		} else
-			e2p_uuid_to_str(sb->s_journal_uuid, buf);
-		fprintf(f, "Journal UUID:             %s\n", buf);
+		fprintf(f, "Journal UUID:             %s\n",
+			e2p_uuid2str(sb->s_journal_uuid));
 		fprintf(f, "Journal inode:            %u\n", sb->s_journal_inum);
 		fprintf(f, "Journal device:	          0x%04x\n", sb->s_journal_dev);
 		fprintf(f, "First orphan inode:       %u\n", sb->s_last_orphan);
 	}
+	if (sb->s_feature_compat & EXT2_FEATURE_COMPAT_DIR_INDEX) {
+		fprintf(f, "Default directory hash:   %s\n",
+			e2p_hash2string(sb->s_def_hash_version));
+		fprintf(f, "Directory Hash Seed:      %s\n",
+			e2p_uuid2str(sb->s_hash_seed));
+	}
 }
 
 void list_super (struct ext2_super_block * s)
