@@ -21,10 +21,3 @@
 extern char const * const sys_errlist[];
 extern const int sys_nerr;
 #endif
-
-/* AIX and Ultrix have standard conforming header files. */
-#if !defined(ultrix) && !defined(_AIX)
-#ifdef __STDC__
-void perror (const char *);
-#endif
-#endif
