@@ -55,7 +55,7 @@ errcode_t ext2fs_new_inode(ext2_filsys fs, ext2_ino_t dir, int mode,
 	i = start_inode;
 
 	do {
-		if (!ext2fs_test_inode_bitmap(map, i))
+		if (!ext2fs_fast_test_inode_bitmap(map, i))
 			break;
 		i++;
 		if (i > fs->super->s_inodes_count)
@@ -87,7 +87,7 @@ errcode_t ext2fs_new_block(ext2_filsys fs, blk_t goal,
 		goal = fs->super->s_first_data_block;
 	i = goal;
 	do {
-		if (!ext2fs_test_block_bitmap(map, i)) {
+		if (!ext2fs_fast_test_block_bitmap(map, i)) {
 			*ret = i;
 			return 0;
 		}
@@ -107,7 +107,6 @@ errcode_t ext2fs_alloc_block(ext2_filsys fs, blk_t goal,
 {
 	errcode_t	retval;
 	blk_t		block;
-	int		group;
 	char		*buf = 0;
 
 	if (!block_buf) {
@@ -132,12 +131,7 @@ errcode_t ext2fs_alloc_block(ext2_filsys fs, blk_t goal,
 	if (retval)
 		goto fail;
 	
-	fs->super->s_free_blocks_count--;
-	group = ext2fs_group_of_blk(fs, block);
-	fs->group_desc[group].bg_free_blocks_count--;
-	ext2fs_mark_block_bitmap(fs->block_map, block);
-	ext2fs_mark_super_dirty(fs);
-	ext2fs_mark_bb_dirty(fs);
+	ext2fs_block_alloc_stats(fs, block, +1);
 	*ret = block;
 	return 0;
 
