@@ -23,28 +23,23 @@
 #include <sys/types.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fsP.h"
 
 /*
  * Helper function for making a badblocks list
  */
-static errcode_t make_badblocks_list(int size, int num, blk_t *list,
-				     ext2_badblocks_list *ret)
+static errcode_t make_u32_list(int size, int num, __u32 *list,
+			       ext2_u32_list *ret)
 {
-	ext2_badblocks_list	bb;
-	errcode_t		retval;
+	ext2_u32_list	bb;
+	errcode_t	retval;
 	
-	retval = ext2fs_get_mem(sizeof(struct ext2_struct_badblocks_list),
+	retval = ext2fs_get_mem(sizeof(struct ext2_struct_u32_list),
 				(void **) &bb);
 	if (retval)
 		return retval;
-	memset(bb, 0, sizeof(struct ext2_struct_badblocks_list));
+	memset(bb, 0, sizeof(struct ext2_struct_u32_list));
 	bb->magic = EXT2_ET_MAGIC_BADBLOCKS_LIST;
 	bb->size = size ? size : 10;
 	bb->num = num;
@@ -62,30 +57,43 @@ static errcode_t make_badblocks_list(int size, int num, blk_t *list,
 }
 	
 
+/*
+ * This procedure creates an empty u32 list.
+ */
+errcode_t ext2fs_u32_list_create(ext2_u32_list *ret, int size)
+{
+	return make_u32_list(size, 0, 0, ret);
+}
+
 /*
  * This procedure creates an empty badblocks list.
  */
 errcode_t ext2fs_badblocks_list_create(ext2_badblocks_list *ret, int size)
 {
-	return make_badblocks_list(size, 0, 0, ret);
+	return make_u32_list(size, 0, 0, (ext2_badblocks_list *) ret);
 }
 
+
 /*
  * This procedure copies a badblocks list
  */
-errcode_t ext2fs_badblocks_copy(ext2_badblocks_list src,
-				ext2_badblocks_list *dest)
+errcode_t ext2fs_u32_copy(ext2_u32_list src, ext2_u32_list *dest)
 {
 	errcode_t	retval;
 	
-	retval = make_badblocks_list(src->size, src->num, src->list,
-				     dest);
+	retval = make_u32_list(src->size, src->num, src->list, dest);
 	if (retval)
 		return retval;
 	(*dest)->badblocks_flags = src->badblocks_flags;
 	return 0;
 }
 
+errcode_t ext2fs_badblocks_copy(ext2_badblocks_list src,
+				ext2_badblocks_list *dest)
+{
+	return ext2fs_u32_copy((ext2_u32_list) src,
+			       (ext2_u32_list *) dest);
+}
 
 /*
  * This procedure frees a badblocks list.
@@ -97,7 +105,7 @@ errcode_t ext2fs_badblocks_copy(ext2_badblocks_list src,
 /*
  * This procedure adds a block to a badblocks list.
  */
-errcode_t ext2fs_badblocks_list_add(ext2_badblocks_list bb, blk_t blk)
+errcode_t ext2fs_u32_list_add(ext2_u32_list bb, __u32 blk)
 {
 	errcode_t	retval;
 	int		i, j;
@@ -106,9 +114,9 @@ errcode_t ext2fs_badblocks_list_add(ext2_badblocks_list bb, blk_t blk)
 	EXT2_CHECK_MAGIC(bb, EXT2_ET_MAGIC_BADBLOCKS_LIST);
 
 	if (bb->num >= bb->size) {
-		old_size = bb->size * sizeof(blk_t);
+		old_size = bb->size * sizeof(__u32);
 		bb->size += 100;
-		retval = ext2fs_resize_mem(old_size, bb->size * sizeof(blk_t),
+		retval = ext2fs_resize_mem(old_size, bb->size * sizeof(__u32),
 					   (void **) &bb->list);
 		if (retval) {
 			bb->size -= 100;
@@ -143,11 +151,18 @@ errcode_t ext2fs_badblocks_list_add(ext2_badblocks_list bb, blk_t blk)
 	return 0;
 }
 
+errcode_t ext2fs_badblocks_list_add(ext2_badblocks_list bb, blk_t blk)
+{
+	return ext2fs_u32_list_add((ext2_u32_list) bb, (__u32) blk);
+}
+
+
+
 /*
  * This procedure tests to see if a particular block is on a badblocks
  * list.
  */
-int ext2fs_badblocks_list_test(ext2_badblocks_list bb, blk_t blk)
+int ext2fs_u32_list_test(ext2_u32_list bb, __u32 blk)
 {
 	int	low, high, mid;
 
@@ -178,15 +193,21 @@ int ext2fs_badblocks_list_test(ext2_badblocks_list bb, blk_t blk)
 	return 0;
 }
 
-errcode_t ext2fs_badblocks_list_iterate_begin(ext2_badblocks_list bb,
-					      ext2_badblocks_iterate *ret)
+int ext2fs_badblocks_list_test(ext2_badblocks_list bb, blk_t blk)
 {
-	ext2_badblocks_iterate iter;
+	return ext2fs_u32_list_test((ext2_u32_list) bb, (__u32) blk);
+}
+
+
+errcode_t ext2fs_u32_list_iterate_begin(ext2_u32_list bb,
+					ext2_u32_iterate *ret)
+{
+	ext2_u32_iterate iter;
 	errcode_t		retval;
 
 	EXT2_CHECK_MAGIC(bb, EXT2_ET_MAGIC_BADBLOCKS_LIST);
 
-	retval = ext2fs_get_mem(sizeof(struct ext2_struct_badblocks_iterate),
+	retval = ext2fs_get_mem(sizeof(struct ext2_struct_u32_iterate),
 			      (void **) &iter);
 	if (retval)
 		return retval;
@@ -198,9 +219,17 @@ errcode_t ext2fs_badblocks_list_iterate_begin(ext2_badblocks_list bb,
 	return 0;
 }
 
-int ext2fs_badblocks_list_iterate(ext2_badblocks_iterate iter, blk_t *blk)
+errcode_t ext2fs_badblocks_list_iterate_begin(ext2_badblocks_list bb,
+					      ext2_badblocks_iterate *ret)
 {
-	ext2_badblocks_list	bb;
+	return ext2fs_u32_list_iterate_begin((ext2_u32_list) bb,
+					      (ext2_u32_iterate *) ret);
+}
+
+
+int ext2fs_u32_list_iterate(ext2_u32_iterate iter, __u32 *blk)
+{
+	ext2_u32_list	bb;
 
 	if (iter->magic != EXT2_ET_MAGIC_BADBLOCKS_ITERATE)
 		return 0;
@@ -218,7 +247,14 @@ int ext2fs_badblocks_list_iterate(ext2_badblocks_iterate iter, blk_t *blk)
 	return 0;
 }
 
-void ext2fs_badblocks_list_iterate_end(ext2_badblocks_iterate iter)
+int ext2fs_badblocks_list_iterate(ext2_badblocks_iterate iter, blk_t *blk)
+{
+	return ext2fs_u32_list_iterate((ext2_u32_iterate) iter,
+				       (__u32 *) blk);
+}
+
+
+void ext2fs_u32_list_iterate_end(ext2_u32_iterate iter)
 {
 	if (!iter || (iter->magic != EXT2_ET_MAGIC_BADBLOCKS_ITERATE))
 		return;
@@ -226,3 +262,28 @@ void ext2fs_badblocks_list_iterate_end(ext2_badblocks_iterate iter)
 	iter->bb = 0;
 	ext2fs_free_mem((void **) &iter);
 }
+
+void ext2fs_badblocks_list_iterate_end(ext2_badblocks_iterate iter)
+{
+	ext2fs_u32_list_iterate_end((ext2_u32_iterate) iter);
+}
+
+
+int ext2fs_u32_list_equal(ext2_u32_list bb1, ext2_u32_list bb2)
+{
+	EXT2_CHECK_MAGIC(bb1, EXT2_ET_MAGIC_BADBLOCKS_LIST);
+	EXT2_CHECK_MAGIC(bb2, EXT2_ET_MAGIC_BADBLOCKS_LIST);
+
+	if (bb1->num != bb2->num)
+		return 0;
+
+	if (memcmp(bb1->list, bb2->list, bb1->num * sizeof(blk_t)) != 0)
+		return 0;
+	return 1;
+}
+
+int ext2fs_badblocks_equal(ext2_badblocks_list bb1, ext2_badblocks_list bb2)
+{
+	return ext2fs_u32_list_equal((ext2_u32_list) bb1,
+				     (ext2_u32_list) bb2);
+}
