@@ -24,12 +24,7 @@
 #include <sys/types.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fs.h"
 
 #if  defined(__linux__)    &&	defined(EXT2_OS_LINUX)
@@ -68,9 +63,11 @@ errcode_t ext2fs_initialize(const char *name, int flags,
 	int		rem;
 	int		overhead = 0;
 	blk_t		group_block;
+	int		ipg;
 	int		i, j;
 	blk_t		numblocks;
 	char		*buf;
+	int		meta_bg_size, meta_bg, has_super, old_desc_blocks;
 
 	if (!param || !param->s_blocks_count)
 		return EXT2_ET_INVALID_ARGUMENT;
@@ -82,7 +79,11 @@ errcode_t ext2fs_initialize(const char *name, int flags,
 	
 	memset(fs, 0, sizeof(struct struct_ext2_filsys));
 	fs->magic = EXT2_ET_MAGIC_EXT2FS_FILSYS;
-	fs->flags = flags | EXT2_FLAG_RW | ext2fs_native_flag();
+	fs->flags = flags | EXT2_FLAG_RW;
+	fs->umask = 022;
+#ifdef WORDS_BIGENDIAN
+	fs->flags |= EXT2_FLAG_SWAP_BYTES;
+#endif
 	retval = manager->open(name, IO_FLAG_RW, &fs->io);
 	if (retval)
 		goto cleanup;
@@ -110,10 +111,10 @@ errcode_t ext2fs_initialize(const char *name, int flags,
 	set_field(s_first_data_block, super->s_log_block_size ? 0 : 1);
 	set_field(s_max_mnt_count, EXT2_DFL_MAX_MNT_COUNT);
 	set_field(s_errors, EXT2_ERRORS_DEFAULT);
-#ifdef EXT2_DYNAMIC_REV
 	set_field(s_feature_compat, 0);
 	set_field(s_feature_incompat, 0);
 	set_field(s_feature_ro_compat, 0);
+	set_field(s_first_meta_bg, 0);
 	if (super->s_feature_incompat & ~EXT2_LIB_FEATURE_INCOMPAT_SUPP)
 		return EXT2_ET_UNSUPP_FEATURE;
 	if (super->s_feature_ro_compat & ~EXT2_LIB_FEATURE_RO_COMPAT_SUPP)
@@ -124,19 +125,20 @@ errcode_t ext2fs_initialize(const char *name, int flags,
 		set_field(s_first_ino, EXT2_GOOD_OLD_FIRST_INO);
 		set_field(s_inode_size, EXT2_GOOD_OLD_INODE_SIZE);
 	}
-#endif
 
 	set_field(s_checkinterval, EXT2_DFL_CHECKINTERVAL);
-	super->s_lastcheck = time(NULL);
+	super->s_mkfs_time = super->s_lastcheck = time(NULL);
 
 	super->s_creator_os = CREATOR_OS;
 
 	fs->blocksize = EXT2_BLOCK_SIZE(super);
 	fs->fragsize = EXT2_FRAG_SIZE(super);
 	frags_per_block = fs->blocksize / fs->fragsize;
-	
-	/* default: (fs->blocksize*8) blocks/group */
-	set_field(s_blocks_per_group, fs->blocksize*8); 
+
+	/* default: (fs->blocksize*8) blocks/group, up to 2^16 (GDT limit) */
+	set_field(s_blocks_per_group, fs->blocksize * 8);
+	if (super->s_blocks_per_group > EXT2_MAX_BLOCKS_PER_GROUP(super))
+		super->s_blocks_per_group = EXT2_MAX_BLOCKS_PER_GROUP(super);
 	super->s_frags_per_group = super->s_blocks_per_group * frags_per_block;
 	
 	super->s_blocks_count = param->s_blocks_count;
@@ -146,6 +148,18 @@ errcode_t ext2fs_initialize(const char *name, int flags,
 		goto cleanup;
 	}
 
+	/*
+	 * If we're creating an external journal device, we don't need
+	 * to bother with the rest.
+	 */
+	if (super->s_feature_incompat &
+	    EXT3_FEATURE_INCOMPAT_JOURNAL_DEV) {
+		fs->group_desc_count = 0;
+		ext2fs_mark_super_dirty(fs);
+		*ret_fs = fs;
+		return 0;
+	}
+
 retry:
 	fs->group_desc_count = (super->s_blocks_count -
 				super->s_first_data_block +
@@ -157,8 +171,8 @@ retry:
 			   EXT2_DESC_PER_BLOCK(super) - 1)
 		/ EXT2_DESC_PER_BLOCK(super);
 
-	/* n.b., fs->blocksize is <= 4096 */
-	set_field(s_inodes_count, super->s_blocks_count/(4096/fs->blocksize));
+	i = fs->blocksize >= 4096 ? 1 : 4096 / fs->blocksize;
+	set_field(s_inodes_count, super->s_blocks_count / i);
 
 	/*
 	 * Make sure we have at least EXT2_FIRST_INO + 1 inodes, so
@@ -171,14 +185,20 @@ retry:
 	 * There should be at least as many inodes as the user
 	 * requested.  Figure out how many inodes per group that
 	 * should be.  But make sure that we don't allocate more than
-	 * one bitmap's worth of inodes
+	 * one bitmap's worth of inodes each group.
 	 */
-	super->s_inodes_per_group = (super->s_inodes_count +
-				     fs->group_desc_count - 1) /
-					     fs->group_desc_count;
-	if (super->s_inodes_per_group > fs->blocksize*8)
-		super->s_inodes_per_group = fs->blocksize*8;
-	
+	ipg = (super->s_inodes_count + fs->group_desc_count - 1) /
+		fs->group_desc_count;
+	if (ipg > fs->blocksize * 8)
+		ipg = fs->blocksize * 8;
+
+	if (ipg > EXT2_MAX_INODES_PER_GROUP(super))
+		ipg = EXT2_MAX_INODES_PER_GROUP(super);
+
+	super->s_inodes_per_group = ipg;
+	if (super->s_inodes_count > ipg * fs->group_desc_count)
+		super->s_inodes_count = ipg * fs->group_desc_count;
+
 	/*
 	 * Make sure the number of inodes per group completely fills
 	 * the inode table blocks in the descriptor.  If not, add some
@@ -273,6 +293,10 @@ retry:
 	 */
 	group_block = super->s_first_data_block;
 	super->s_free_blocks_count = 0;
+	if (fs->super->s_feature_incompat & EXT2_FEATURE_INCOMPAT_META_BG)
+		old_desc_blocks = fs->super->s_first_meta_bg;
+	else
+		old_desc_blocks = fs->desc_blocks;
 	for (i = 0; i < fs->group_desc_count; i++) {
 		if (i == fs->group_desc_count-1) {
 			numblocks = (fs->super->s_blocks_count -
@@ -283,11 +307,35 @@ retry:
 		} else
 			numblocks = fs->super->s_blocks_per_group;
 
-		if (ext2fs_bg_has_super(fs, i)) {
-			for (j=0; j < fs->desc_blocks+1; j++)
+		has_super = ext2fs_bg_has_super(fs, i);
+
+		if (has_super) {
+			ext2fs_mark_block_bitmap(fs->block_map, group_block);
+			numblocks--;
+		}
+		meta_bg_size = (fs->blocksize /
+				sizeof (struct ext2_group_desc));
+		meta_bg = i / meta_bg_size;
+
+		if (!(fs->super->s_feature_incompat &
+		      EXT2_FEATURE_INCOMPAT_META_BG) ||
+		    (meta_bg < fs->super->s_first_meta_bg)) {
+			if (has_super) {
+				for (j=0; j < old_desc_blocks; j++)
+					ext2fs_mark_block_bitmap(fs->block_map,
+							 group_block + j + 1);
+				numblocks -= old_desc_blocks;
+			}
+		} else {
+			if (has_super)
+				has_super = 1;
+			if (((i % meta_bg_size) == 0) ||
+			    ((i % meta_bg_size) == 1) ||
+			    ((i % meta_bg_size) == (meta_bg_size-1))) {
 				ext2fs_mark_block_bitmap(fs->block_map,
-							 group_block + j);
-			numblocks -= 1 + fs->desc_blocks;
+					 group_block + has_super);
+				numblocks--;
+			}
 		}
 		
 		numblocks -= 2 + fs->inode_blocks_per_group;
@@ -313,6 +361,3 @@ cleanup:
 	ext2fs_free(fs);
 	return retval;
 }
-	
-
-
