@@ -21,21 +21,17 @@
 #include <sys/types.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fsP.h"
+#include "e2image.h"
 
 struct ext2_struct_inode_scan {
 	errcode_t		magic;
 	ext2_filsys		fs;
-	ino_t			current_inode;
+	ext2_ino_t		current_inode;
 	blk_t			current_block;
 	dgrp_t			current_group;
-	ino_t			inodes_left;
+	ext2_ino_t		inodes_left;
 	blk_t			blocks_left;
 	dgrp_t			groups_left;
 	blk_t			inode_buffer_blocks;
@@ -54,10 +50,25 @@ struct ext2_struct_inode_scan {
 	int			reserved[6];
 };
 
+/*
+ * This routine flushes the icache, if it exists.
+ */
+errcode_t ext2fs_flush_icache(ext2_filsys fs)
+{
+	int	i;
+	
+	if (!fs->icache)
+		return 0;
+
+	for (i=0; i < fs->icache->cache_size; i++)
+		fs->icache->cache[i].ino = 0;
+
+	return 0;
+}
+
 static errcode_t create_icache(ext2_filsys fs)
 {
 	errcode_t	retval;
-	int		i;
 	
 	if (fs->icache)
 		return 0;
@@ -84,8 +95,7 @@ static errcode_t create_icache(ext2_filsys fs)
 		ext2fs_free_mem((void **) &fs->icache);
 		return retval;
 	}
-	for (i=0; i < fs->icache->cache_size; i++)
-		fs->icache->cache[i].ino = 0;
+	ext2fs_flush_icache(fs);
 	return 0;
 }
 
@@ -94,7 +104,7 @@ errcode_t ext2fs_open_inode_scan(ext2_filsys fs, int buffer_blocks,
 {
 	ext2_inode_scan	scan;
 	errcode_t	retval;
-	errcode_t (*save_get_blocks)(ext2_filsys f, ino_t ino, blk_t *blocks);
+	errcode_t (*save_get_blocks)(ext2_filsys f, ext2_ino_t ino, blk_t *blocks);
 
 	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);
 
@@ -111,7 +121,7 @@ errcode_t ext2fs_open_inode_scan(ext2_filsys fs, int buffer_blocks,
 		fs->get_blocks = 0;
 		retval = ext2fs_read_bb_inode(fs, &fs->badblocks);
 		if (retval && fs->badblocks) {
-			badblocks_list_free(fs->badblocks);
+			ext2fs_badblocks_list_free(fs->badblocks);
 			fs->badblocks = 0;
 		}
 		fs->get_blocks = save_get_blocks;
@@ -367,7 +377,7 @@ static inline int is_empty_scan(ext2_inode_scan scan)
 }
 #endif
 
-errcode_t ext2fs_get_next_inode(ext2_inode_scan scan, ino_t *ino,
+errcode_t ext2fs_get_next_inode(ext2_inode_scan scan, ext2_ino_t *ino,
 				struct ext2_inode *inode)
 {
 	errcode_t	retval;
@@ -435,21 +445,25 @@ errcode_t ext2fs_get_next_inode(ext2_inode_scan scan, ino_t *ino,
 		scan->ptr += scan->inode_size - extra_bytes;
 		scan->bytes_left -= scan->inode_size - extra_bytes;
 
+#ifdef EXT2FS_ENABLE_SWAPFS
 		if ((scan->fs->flags & EXT2_FLAG_SWAP_BYTES) ||
 		    (scan->fs->flags & EXT2_FLAG_SWAP_BYTES_READ))
 			ext2fs_swap_inode(scan->fs, inode,
 				 (struct ext2_inode *) scan->temp_buffer, 0);
 		else
+#endif
 			*inode = *((struct ext2_inode *) scan->temp_buffer);
 		if (scan->scan_flags & EXT2_SF_BAD_EXTRA_BYTES)
 			retval = EXT2_ET_BAD_BLOCK_IN_INODE_TABLE;
 		scan->scan_flags &= ~EXT2_SF_BAD_EXTRA_BYTES;
 	} else {
+#ifdef EXT2FS_ENABLE_SWAPFS
 		if ((scan->fs->flags & EXT2_FLAG_SWAP_BYTES) ||
 		    (scan->fs->flags & EXT2_FLAG_SWAP_BYTES_READ))
 			ext2fs_swap_inode(scan->fs, inode,
 				 (struct ext2_inode *) scan->ptr, 0);
 		else
+#endif
 			*inode = *((struct ext2_inode *) scan->ptr);
 		scan->ptr += scan->inode_size;
 		scan->bytes_left -= scan->inode_size;
@@ -466,13 +480,13 @@ errcode_t ext2fs_get_next_inode(ext2_inode_scan scan, ino_t *ino,
 /*
  * Functions to read and write a single inode.
  */
-errcode_t ext2fs_read_inode (ext2_filsys fs, ino_t ino,
+errcode_t ext2fs_read_inode (ext2_filsys fs, ext2_ino_t ino,
 			     struct ext2_inode * inode)
 {
 	unsigned long 	group, block, block_nr, offset;
 	char 		*ptr;
 	errcode_t	retval;
-	int 		clen, length, i;
+	int 		clen, length, i, inodes_per_block;
 
 	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);
 
@@ -497,13 +511,22 @@ errcode_t ext2fs_read_inode (ext2_filsys fs, ino_t ino,
 	}
 	if ((ino == 0) || (ino > fs->super->s_inodes_count))
 		return EXT2_ET_BAD_INODE_NUM;
-	group = (ino - 1) / EXT2_INODES_PER_GROUP(fs->super);
-	offset = ((ino - 1) % EXT2_INODES_PER_GROUP(fs->super)) *
-		EXT2_INODE_SIZE(fs->super);
-	block = offset >> EXT2_BLOCK_SIZE_BITS(fs->super);
-	if (!fs->group_desc[(unsigned)group].bg_inode_table)
-		return EXT2_ET_MISSING_INODE_TABLE;
-	block_nr = fs->group_desc[(unsigned)group].bg_inode_table + block;
+	if (fs->flags & EXT2_FLAG_IMAGE_FILE) {
+		inodes_per_block = fs->blocksize / EXT2_INODE_SIZE(fs->super);
+		block_nr = fs->image_header->offset_inode / fs->blocksize;
+		block_nr += (ino - 1) / inodes_per_block;
+		offset = ((ino - 1) % inodes_per_block) *
+			EXT2_INODE_SIZE(fs->super);
+	} else {
+		group = (ino - 1) / EXT2_INODES_PER_GROUP(fs->super);
+		offset = ((ino - 1) % EXT2_INODES_PER_GROUP(fs->super)) *
+			EXT2_INODE_SIZE(fs->super);
+		block = offset >> EXT2_BLOCK_SIZE_BITS(fs->super);
+		if (!fs->group_desc[(unsigned)group].bg_inode_table)
+			return EXT2_ET_MISSING_INODE_TABLE;
+		block_nr = fs->group_desc[(unsigned)group].bg_inode_table + 
+			block;
+	}
 	if (block_nr != fs->icache->buffer_blk) {
 		retval = io_channel_read_blk(fs->io, block_nr, 1,
 					     fs->icache->buffer);
@@ -537,9 +560,11 @@ errcode_t ext2fs_read_inode (ext2_filsys fs, ino_t ino,
 	} else
 		memcpy((char *) inode, ptr, length);
 	
+#ifdef EXT2FS_ENABLE_SWAPFS
 	if ((fs->flags & EXT2_FLAG_SWAP_BYTES) ||
 	    (fs->flags & EXT2_FLAG_SWAP_BYTES_READ))
 		ext2fs_swap_inode(fs, inode, inode, 0);
+#endif
 
 	/* Update the inode cache */
 	fs->icache->cache_last = (fs->icache->cache_last + 1) %
@@ -550,7 +575,7 @@ errcode_t ext2fs_read_inode (ext2_filsys fs, ino_t ino,
 	return 0;
 }
 
-errcode_t ext2fs_write_inode(ext2_filsys fs, ino_t ino,
+errcode_t ext2fs_write_inode(ext2_filsys fs, ext2_ino_t ino,
 			     struct ext2_inode * inode)
 {
 	unsigned long group, block, block_nr, offset;
@@ -588,10 +613,12 @@ errcode_t ext2fs_write_inode(ext2_filsys fs, ino_t ino,
 	if ((ino == 0) || (ino > fs->super->s_inodes_count))
 		return EXT2_ET_BAD_INODE_NUM;
 
+#ifdef EXT2FS_ENABLE_SWAPFS
 	if ((fs->flags & EXT2_FLAG_SWAP_BYTES) ||
 	    (fs->flags & EXT2_FLAG_SWAP_BYTES_WRITE))
 		ext2fs_swap_inode(fs, &temp_inode, inode, 1);
 	else
+#endif
 		memcpy(&temp_inode, inode, sizeof(struct ext2_inode));
 	
 	group = (ino - 1) / EXT2_INODES_PER_GROUP(fs->super);
@@ -649,7 +676,7 @@ errcode_t ext2fs_write_inode(ext2_filsys fs, ino_t ino,
 	return 0;
 }
 
-errcode_t ext2fs_get_blocks(ext2_filsys fs, ino_t ino, blk_t *blocks)
+errcode_t ext2fs_get_blocks(ext2_filsys fs, ext2_ino_t ino, blk_t *blocks)
 {
 	struct ext2_inode	inode;
 	int			i;
@@ -672,7 +699,7 @@ errcode_t ext2fs_get_blocks(ext2_filsys fs, ino_t ino, blk_t *blocks)
 	return 0;
 }
 
-errcode_t ext2fs_check_directory(ext2_filsys fs, ino_t ino)
+errcode_t ext2fs_check_directory(ext2_filsys fs, ext2_ino_t ino)
 {
 	struct	ext2_inode	inode;
 	errcode_t		retval;
