@@ -23,28 +23,22 @@
 #include <sys/types.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fs.h"
 
 #ifndef EXT2_FT_DIR
 #define EXT2_FT_DIR		2
 #endif
 
-errcode_t ext2fs_mkdir(ext2_filsys fs, ino_t parent, ino_t inum,
+errcode_t ext2fs_mkdir(ext2_filsys fs, ext2_ino_t parent, ext2_ino_t inum,
 		       const char *name)
 {
 	errcode_t		retval;
-	struct ext2_inode	inode;
-	ino_t			ino = inum;
-	ino_t			scratch_ino;
+	struct ext2_inode	parent_inode, inode;
+	ext2_ino_t		ino = inum;
+	ext2_ino_t		scratch_ino;
 	blk_t			blk;
 	char			*block = 0;
-	int			group;
 
 	EXT2_CHECK_MAGIC(fs, EXT2_ET_MAGIC_EXT2FS_FILSYS);
 
@@ -72,11 +66,21 @@ errcode_t ext2fs_mkdir(ext2_filsys fs, ino_t parent, ino_t inum,
 	if (retval)
 		goto cleanup;
 
+	/*
+	 * Get the parent's inode, if necessary
+	 */
+	if (parent != ino) {
+		retval = ext2fs_read_inode(fs, parent, &parent_inode);
+		if (retval)
+			goto cleanup;
+	} else
+		memset(&parent_inode, 0, sizeof(parent_inode));
+
 	/*
 	 * Create the inode structure....
 	 */
 	memset(&inode, 0, sizeof(struct ext2_inode));
-	inode.i_mode = LINUX_S_IFDIR | 0755;
+	inode.i_mode = LINUX_S_IFDIR | (0777 & ~fs->umask);
 	inode.i_uid = inode.i_gid = 0;
 	inode.i_blocks = fs->blocksize / 512;
 	inode.i_block[0] = blk;
@@ -94,19 +98,6 @@ errcode_t ext2fs_mkdir(ext2_filsys fs, ino_t parent, ino_t inum,
 	if (retval)
 		goto cleanup;
 
-	/*
-	 * Update parent inode's counts
-	 */
-	if (parent != ino) {
-		retval = ext2fs_read_inode(fs, parent, &inode);
-		if (retval)
-			goto cleanup;
-		inode.i_links_count++;
-		retval = ext2fs_write_inode(fs, parent, &inode);
-		if (retval)
-			goto cleanup;
-	}
-	
 	/*
 	 * Link the directory into the filesystem hierarchy
 	 */
@@ -125,23 +116,22 @@ errcode_t ext2fs_mkdir(ext2_filsys fs, ino_t parent, ino_t inum,
 			goto cleanup;
 	}
 
+	/*
+	 * Update parent inode's counts
+	 */
+	if (parent != ino) {
+		parent_inode.i_links_count++;
+		retval = ext2fs_write_inode(fs, parent, &parent_inode);
+		if (retval)
+			goto cleanup;
+	}
+	
 	/*
 	 * Update accounting....
 	 */
-	ext2fs_mark_block_bitmap(fs->block_map, blk);
-	ext2fs_mark_bb_dirty(fs);
-	ext2fs_mark_inode_bitmap(fs->inode_map, ino);
-	ext2fs_mark_ib_dirty(fs);
+	ext2fs_block_alloc_stats(fs, blk, +1);
+	ext2fs_inode_alloc_stats2(fs, ino, +1, 1);
 
-	group = ext2fs_group_of_blk(fs, blk);
-	fs->group_desc[group].bg_free_blocks_count--;
-	group = ext2fs_group_of_ino(fs, ino);
-	fs->group_desc[group].bg_free_inodes_count--;
-	fs->group_desc[group].bg_used_dirs_count++;
-	fs->super->s_free_blocks_count--;
-	fs->super->s_free_inodes_count--;
-	ext2fs_mark_super_dirty(fs);
-	
 cleanup:
 	if (block)
 		ext2fs_free_mem((void **) &block);
