@@ -24,9 +24,41 @@
 #endif
 
 #include "ext2_fs.h"
+
+
 #include "ext2fs.h"
 #include "e2image.h"
 
+blk_t ext2fs_descriptor_block_loc(ext2_filsys fs, blk_t group_block, dgrp_t i)
+{
+	int	bg;
+	int	has_super = 0;
+	int	ret_blk;
+
+	if (!(fs->super->s_feature_incompat & EXT2_FEATURE_INCOMPAT_META_BG) ||
+	    (i < fs->super->s_first_meta_bg))
+		return (group_block + i + 1);
+
+	bg = (fs->blocksize / sizeof (struct ext2_group_desc)) * i;
+	if (ext2fs_bg_has_super(fs, bg))
+		has_super = 1;
+	ret_blk = (fs->super->s_first_data_block + has_super + 
+		   (bg * fs->super->s_blocks_per_group));
+	/*
+	 * If group_block is not the normal value, we're trying to use
+	 * the backup group descriptors and superblock --- so use the
+	 * alternate location of the second block group in the
+	 * metablock group.  Ideally we should be testing each bg
+	 * descriptor block individually for correctness, but we don't
+	 * have the infrastructure in place to do that.
+	 */
+	if (group_block != fs->super->s_first_data_block &&
+	    ((ret_blk + fs->super->s_blocks_per_group) <
+	     fs->super->s_blocks_count))
+		ret_blk += fs->super->s_blocks_per_group;
+	return ret_blk;
+}
+
 /*
  *  Note: if superblock is non-zero, block-size must also be non-zero.
  * 	Superblock and block_size can be zero to use the default size.
@@ -43,8 +75,8 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 {
 	ext2_filsys	fs;
 	errcode_t	retval;
-	int		i, j, groups_per_block;
-	blk_t		group_block;
+	int		i, j, groups_per_block, blocks_per_group;
+	blk_t		group_block, blk;
 	char		*dest;
 	struct ext2_group_desc *gdp;
 	
@@ -102,7 +134,7 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 			goto cleanup;
 		}
 		io_channel_set_blksize(fs->io, block_size);
-		group_block = superblock + 1;
+		group_block = superblock;
 		fs->orig_super = 0;
 	} else {
 		io_channel_set_blksize(fs->io, SUPERBLOCK_OFFSET);
@@ -196,14 +228,16 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 	/*
 	 * Read group descriptors
 	 */
-	if ((EXT2_BLOCKS_PER_GROUP(fs->super)) == 0) {
+	blocks_per_group = EXT2_BLOCKS_PER_GROUP(fs->super);
+	if (blocks_per_group == 0 ||
+	    blocks_per_group > EXT2_MAX_BLOCKS_PER_GROUP(fs->super) ||
+	    fs->inode_blocks_per_group > EXT2_MAX_INODES_PER_GROUP(fs->super)) {
 		retval = EXT2_ET_CORRUPT_SUPERBLOCK;
 		goto cleanup;
 	}
 	fs->group_desc_count = (fs->super->s_blocks_count -
 				fs->super->s_first_data_block +
-				EXT2_BLOCKS_PER_GROUP(fs->super) - 1)
-		/ EXT2_BLOCKS_PER_GROUP(fs->super);
+				blocks_per_group - 1) / blocks_per_group;
 	fs->desc_blocks = (fs->group_desc_count +
 			   EXT2_DESC_PER_BLOCK(fs->super) - 1)
 		/ EXT2_DESC_PER_BLOCK(fs->super);
@@ -212,18 +246,17 @@ errcode_t ext2fs_open(const char *name, int flags, int superblock,
 	if (retval)
 		goto cleanup;
 	if (!group_block)
-		group_block = fs->super->s_first_data_block + 1;
+		group_block = fs->super->s_first_data_block;
 	dest = (char *) fs->group_desc;
+	groups_per_block = fs->blocksize / sizeof(struct ext2_group_desc);
 	for (i=0 ; i < fs->desc_blocks; i++) {
-		retval = io_channel_read_blk(fs->io, group_block, 1, dest);
+		blk = ext2fs_descriptor_block_loc(fs, group_block, i);
+		retval = io_channel_read_blk(fs->io, blk, 1, dest);
 		if (retval)
 			goto cleanup;
-		group_block++;
 #ifdef EXT2FS_ENABLE_SWAPFS
 		if (fs->flags & EXT2_FLAG_SWAP_BYTES) {
 			gdp = (struct ext2_group_desc *) dest;
-			groups_per_block = fs->blocksize /
-				sizeof(struct ext2_group_desc);
 			for (j=0; j < groups_per_block; j++)
 				ext2fs_swap_group_desc(gdp++);
 		}
