@@ -22,16 +22,11 @@
 #include <errno.h>
 #endif
 
-#if EXT2_FLAT_INCLUDES
 #include "ext2_fs.h"
-#else
-#include <linux/ext2_fs.h>
-#endif
-
 #include "ext2fs.h"
 
 blk_t test1[] = { 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 0 };
-blk_t test2[] = { 10, 9, 8, 7, 6, 5, 4, 3, 3, 2, 1, 0 };
+blk_t test2[] = { 11, 10, 9, 8, 7, 6, 5, 4, 3, 3, 2, 1 };
 blk_t test3[] = { 3, 1, 4, 5, 9, 2, 7, 10, 5, 6, 10, 8, 0 };
 blk_t test4[] = { 20, 50, 12, 17, 13, 2, 66, 23, 56, 0 };
 blk_t test4a[] = {
@@ -123,45 +118,120 @@ static void validate_test_seq(badblocks_list bb, blk_t *vec)
 	}
 }
 
-int main(int argc, char *argv)
+int file_test(badblocks_list bb)
 {
-	badblocks_list bb;
+	char	tmp_filename[20] = "#testXXXXXX";
+	badblocks_list new_bb = 0;
+	errcode_t	retval;
+	FILE	*f;
+
+	mktemp(tmp_filename);
+
+	unlink(tmp_filename);
+	f = fopen(tmp_filename, "w");
+	if (!f) {
+		fprintf(stderr, "Error opening temp file %s: %s\n",
+			tmp_filename, error_message(errno));
+		return 1;
+	}
+	retval = ext2fs_write_bb_FILE(bb, 0, f);
+	if (retval) {
+		com_err("file_test", retval, "while writing bad blocks");
+		return 1;
+	}
+	fclose(f);
+
+	f = fopen(tmp_filename, "r");
+	if (!f) {
+		fprintf(stderr, "Error re-opening temp file %s: %s\n",
+			tmp_filename, error_message(errno));
+		return 1;
+	}
+	retval = ext2fs_read_bb_FILE2(0, f, &new_bb, 0, 0);
+	if (retval) {
+		com_err("file_test", retval, "while reading bad blocks");
+		return 1;
+	}
+	fclose(f);
+
+	if (ext2fs_badblocks_equal(bb, new_bb)) {
+		printf("Block bitmap matched after reading and writing.\n");
+	} else {
+		printf("Block bitmap NOT matched.\n");
+		test_fail++;
+	}
+	return 0;
+}
+
+
+int main(int argc, char **argv)
+{
+	badblocks_list bb1, bb2, bb3, bb4;
+	int	equal;
 	errcode_t	retval;
 
+	bb1 = bb2 = bb3 = bb4 = 0;
+
 	printf("test1: ");
-	retval = create_test_list(test1, &bb);
-	if (retval == 0) {
-		print_list(bb, 1);
-		ext2fs_badblocks_list_free(bb);
-	}
+	retval = create_test_list(test1, &bb1);
+	if (retval == 0)
+		print_list(bb1, 1);
 	printf("\n");
 	
 	printf("test2: ");
-	retval = create_test_list(test2, &bb);
-	if (retval == 0) {
-		print_list(bb, 1);
-		ext2fs_badblocks_list_free(bb);
-	}
+	retval = create_test_list(test2, &bb2);
+	if (retval == 0)
+		print_list(bb2, 1);
 	printf("\n");
 
 	printf("test3: ");
-	retval = create_test_list(test3, &bb);
-	if (retval == 0) {
-		print_list(bb, 1);
-		ext2fs_badblocks_list_free(bb);
-	}
+	retval = create_test_list(test3, &bb3);
+	if (retval == 0)
+		print_list(bb3, 1);
 	printf("\n");
 	
 	printf("test4: ");
-	retval = create_test_list(test4, &bb);
+	retval = create_test_list(test4, &bb4);
 	if (retval == 0) {
-		print_list(bb, 0);
+		print_list(bb4, 0);
 		printf("\n");
-		validate_test_seq(bb, test4a);
-		ext2fs_badblocks_list_free(bb);
+		validate_test_seq(bb4, test4a);
 	}
 	printf("\n");
+
+	if (bb1 && bb2 && bb3 && bb4) {
+		printf("Comparison tests:\n");
+		equal = ext2fs_badblocks_equal(bb1, bb2);
+		printf("bb1 and bb2 are %sequal.\n", equal ? "" : "NOT "); 
+		if (equal)
+			test_fail++;
+
+		equal = ext2fs_badblocks_equal(bb1, bb3);
+		printf("bb1 and bb3 are %sequal.\n", equal ? "" : "NOT "); 
+		if (!equal)
+			test_fail++;
+		
+		equal = ext2fs_badblocks_equal(bb1, bb4);
+		printf("bb1 and bb4 are %sequal.\n", equal ? "" : "NOT "); 
+		if (equal)
+			test_fail++;
+		printf("\n");
+	}
+	
 	if (test_fail == 0)
 		printf("ext2fs library badblocks tests checks out OK!\n");
+
+	file_test(bb4);
+	
+	if (bb1)
+		ext2fs_badblocks_list_free(bb1);
+	if (bb2)
+		ext2fs_badblocks_list_free(bb2);
+	if (bb3)
+		ext2fs_badblocks_list_free(bb3);
+	if (bb4)
+		ext2fs_badblocks_list_free(bb4);
+
 	return test_fail;
+
 }
