@@ -763,7 +763,8 @@ int main (int argc, char ** argv)
 		check_mount(device_name);
 	
 	dev = open (device_name, O_RDWR);
-	if ((dev == -1) && ((errno == EPERM) || (errno == EACCES)) &&
+	if ((dev == -1) && ((errno == EPERM) || (errno == EACCES) ||
+			    (errno == EROFS)) &&
 	    (w_flag == 0))
 		dev = open(device_name, O_RDONLY);
 	if (dev == -1) {
@@ -784,7 +785,7 @@ int main (int argc, char ** argv)
 		}
 	} else
 		host_dev = dev;
-	if (input_file)
+	if (input_file) {
 		if (strcmp (input_file, "-") == 0)
 			in = stdin;
 		else {
@@ -797,6 +798,7 @@ int main (int argc, char ** argv)
 				exit (1);
 			}
 		}
+	}
 	if (output_file && strcmp (output_file, "-") != 0)
 	{
 		out = fopen (output_file, "w");
