@@ -240,7 +240,7 @@ static void mark_table_blocks(ext2_filsys fs)
 	 */
 	for (j = 0; j < fs->desc_blocks; j++) {
 		ext2fs_mark_block_bitmap(meta_block_map,
-					 block + j + 1);
+			 ext2fs_descriptor_block_loc(fs, block, j));
 	}
 
 	for (i = 0; i < fs->group_desc_count; i++) {
@@ -388,6 +388,8 @@ static void write_raw_image_file(ext2_filsys fs, int fd)
 	stashed_inode = &inode;
 	while (1) {
 		retval = ext2fs_get_next_inode(scan, &ino, &inode);
+		if (retval == EXT2_ET_BAD_BLOCK_IN_INODE_TABLE)
+			continue;
 		if (retval) {
 			com_err(program_name, retval,
 				_("while getting next inode"));
@@ -395,8 +397,13 @@ static void write_raw_image_file(ext2_filsys fs, int fd)
 		}
 		if (ino == 0)
 			break;
-		if (!inode.i_links_count ||
-		    !ext2fs_inode_has_valid_blocks(&inode))
+		if (!inode.i_links_count)
+			continue;
+		if (inode.i_file_acl) {
+			ext2fs_mark_block_bitmap(meta_block_map,
+						 inode.i_file_acl);
+		}
+		if (!ext2fs_inode_has_valid_blocks(&inode))
 			continue;
 		
 		stashed_ino = ino;
@@ -423,10 +430,6 @@ static void write_raw_image_file(ext2_filsys fs, int fd)
 					exit(1);
 				}
 			}
-			if (inode.i_file_acl) {
-				ext2fs_mark_block_bitmap(meta_block_map,
-							 inode.i_file_acl);
-			}
 		}
 	}
 	use_inode_shortcuts(fs, 0);
