@@ -37,6 +37,7 @@
 extern char *ext2fs_find_block_device(dev_t device);
 
 #define PROC_PARTITIONS "/proc/partitions"
+#define PROC_EVMS_VOLUMES "/proc/evms/volumes"
 #define DEVLABELDIR	"/dev"
 #define VG_DIR          "/proc/lvm/VGs"
 
@@ -59,6 +60,40 @@ struct xfs_super_block {
 	unsigned char	s_fname[12];
 };
 
+struct reiserfs_super_block
+{
+ 	/* Following entries are based on reiserfsutils 3.6.3 
+	 * (Copyright Hans Reiser) since Linux kernel headers
+	 * (2.4.18) seemed not up-to-date. */
+	unsigned char	s_dummy1[52];
+	unsigned char	s_magic[10];
+	unsigned char	s_dummy2[10];
+	unsigned char	s_version[2];
+	unsigned char	s_dummy3[10];
+	unsigned char	s_uuid[16];
+	unsigned char	s_label[16];
+	unsigned char	s_unused[88];
+};
+
+#define REISER2FS_SUPER_MAGIC_STRING "ReIsEr2Fs" /* v. 3.6 */
+#define REISER3FS_SUPER_MAGIC_STRING "ReIsEr3Fs" /* Journal Relocation */
+#define REISERFS_DISK_OFFSET_IN_BYTES (64 * 1024)
+/* the spot for the super in versions 3.5 - 3.5.10 (inclusive) - 
+ * We'll use it in case volume has been converted. */
+#define REISERFS_OLD_DISK_OFFSET_IN_BYTES (8 * 1024)
+#define reiserversion(s)	((unsigned) (s).s_version[0] + (((unsigned) (s).s_version[1]) << 8))
+
+/* We're checking for ReiserFS v. 3.6 and RJ 3.6 SB */
+static int
+reiser_supports_uuid (struct reiserfs_super_block *sb)
+{
+	return (strncmp(sb->s_magic, REISER2FS_SUPER_MAGIC_STRING,
+			    strlen (REISER2FS_SUPER_MAGIC_STRING)) == 0)
+		|| (strncmp(sb->s_magic, REISER3FS_SUPER_MAGIC_STRING,
+			    strlen (REISER3FS_SUPER_MAGIC_STRING)) == 0
+				&& reiserversion(*sb) == 2);
+}
+
 static struct uuidCache_s {
 	struct uuidCache_s *next;
 	char uuid[16];
@@ -76,17 +111,18 @@ char *string_copy(const char *s)
 	return ret;
 }
 
-/* for now, only ext2 and xfs are supported */
+/* for now, only ext2, ext3, xfs and ReiserFS are supported */
 static int
 get_label_uuid(const char *device, char **label, char *uuid) {
 
-	/* start with ext2 and xfs tests, taken from mount_guess_fstype */
+	/* start with ext2/3, xfs and ReiserFS tests, taken from mount_guess_fstype */
 	/* should merge these later */
 	int fd;
 	size_t label_size;
 	unsigned char *sb_uuid = 0, *sb_label = 0;
 	struct ext2_super_block e2sb;
 	struct xfs_super_block xfsb;
+	struct reiserfs_super_block rfsb;
 
 	fd = open(device, O_RDONLY);
 	if (fd < 0)
@@ -104,6 +140,17 @@ get_label_uuid(const char *device, char **label, char *uuid) {
 		sb_uuid = xfsb.s_uuid;
 		sb_label = xfsb.s_fname;
 		label_size = sizeof(xfsb.s_fname);
+	} else if ((lseek(fd, REISERFS_OLD_DISK_OFFSET_IN_BYTES, SEEK_SET)
+			    == REISERFS_OLD_DISK_OFFSET_IN_BYTES
+	    && read(fd, (char *) &rfsb, sizeof(rfsb)) == sizeof(rfsb)
+	    && (reiser_supports_uuid(&rfsb)))
+	    || (lseek(fd, REISERFS_DISK_OFFSET_IN_BYTES, SEEK_SET)
+			    == REISERFS_DISK_OFFSET_IN_BYTES
+	    && read(fd, (char *) &rfsb, sizeof(rfsb)) == sizeof(rfsb)
+	    && (reiser_supports_uuid(&rfsb)))) {
+		sb_uuid = rfsb.s_uuid;
+		sb_label = rfsb.s_label;
+		label_size = sizeof(rfsb.s_label);
 	} else {
 		close(fd);
 		return 1;
@@ -119,6 +166,8 @@ get_label_uuid(const char *device, char **label, char *uuid) {
 	return 0;
 }
 
+#define CBBUF  (16 * 1024)
+
 static void
 uuidcache_addentry(char *device, char *label, char *uuid) {
 	struct uuidCache_s *last;
@@ -166,7 +215,7 @@ static void init_lvm(void)
 
 		lv_list = opendir(vdirname);
 		free(vdirname);
-		if (lv_list != NULL)
+		if (lv_list == NULL)
 			return;
 
 		while ((lv_iter = readdir(lv_list)) != 0) {
@@ -197,7 +246,8 @@ static void init_lvm(void)
 #endif
 
 static void
-uuidcache_init(void) {
+read_partitions(void)
+{
 	char line[100];
 	char *s;
 	int ma, mi, sz;
@@ -209,18 +259,16 @@ uuidcache_init(void) {
 	struct stat statbuf;
 	int firstPass;
 	int handleOnFirst;
+	char *iobuf;
 
-	if (uuidCache)
-		return;
-
-#ifdef VG_DIR
-	init_lvm();
-#endif
-	
 	procpt = fopen(PROC_PARTITIONS, "r");
 	if (!procpt)
 		return;
 
+	iobuf = (char *)malloc(CBBUF);
+	if (iobuf)
+		setvbuf(procpt, iobuf, _IOFBF, CBBUF);
+
 	for (firstPass = 1; firstPass >= 0; firstPass--) {
 	    fseek(procpt, 0, SEEK_SET);
 
@@ -241,9 +289,10 @@ uuidcache_init(void) {
 		/* skip entire disk (minor 0, 64, ... on ide;
 		   0, 16, ... on sd) */
 		/* heuristic: partition name ends in a digit */
+		/* OR partition name starts with 'lvm' */
 
 		for(s = ptname; *s; s++);
-		if (isdigit(s[-1])) {
+		if (isdigit(s[-1]) || !strncmp(ptname, "lvm", 3)) {
 			/*
 			 * We first look in /dev for the device, but
 			 * if we don't find it, or if the stat
@@ -259,6 +308,10 @@ uuidcache_init(void) {
 				devname = string_copy(device);
 			if (!devname)
 				continue;
+#ifdef DEBUG
+			printf("Checking partition %s (%d, %d)\n",
+			       devname, ma, mi);
+#endif
 			if (!get_label_uuid(devname, &label, uuid))
 				uuidcache_addentry(devname, label, uuid);
 			else
@@ -268,6 +321,65 @@ uuidcache_init(void) {
 	}
 
 	fclose(procpt);
+	if (iobuf)
+		free(iobuf);
+}
+
+static void
+read_evms(void)
+{
+	char line[100];
+	int ma, mi, sz;
+	FILE *procpt;
+	char uuid[16], *label, *devname;
+	char device[110];
+	dev_t	dev;
+	struct stat statbuf;
+
+	procpt = fopen(PROC_EVMS_VOLUMES, "r");
+	if (!procpt)
+		return;
+	while (fgets(line, sizeof(line), procpt)) {
+		if (sscanf (line, " %d %d %d %*s %*s %[^\n ]",
+			    &ma, &mi, &sz, device) != 4)
+			continue;
+
+		/*
+		 * We first look for the device in the named location,
+		 * but if we don't find it, or if the stat information
+		 * doesn't check out, we use ext2fs_find_block_device
+		 * to find it.
+		 */
+		dev = makedev(ma, mi);
+		if ((stat(device, &statbuf) < 0) || (statbuf.st_rdev != dev)) {
+			devname = ext2fs_find_block_device(dev);
+		} else
+			devname = string_copy(device);
+		if (!devname)
+			continue;
+#ifdef DEBUG
+		printf("Checking partition %s (%d, %d)\n",
+		       devname, ma, mi);
+#endif
+		if (!get_label_uuid(devname, &label, uuid))
+			uuidcache_addentry(devname, label, uuid);
+		else
+			free(devname);
+	}
+	fclose(procpt);
+}
+
+static void
+uuidcache_init(void)
+{
+	if (uuidCache)
+		return;
+
+#ifdef VG_DIR
+	init_lvm();
+#endif
+	read_evms();
+	read_partitions();
 }
 
 #define UUID   1
@@ -328,7 +440,7 @@ get_spec_by_uuid(const char *s)
 	return get_spec_by_x(UUID, uuid);
 
  bad_uuid:
-	fprintf(stderr, _("WARNING: %s: bad UUID"), s);
+	fprintf(stderr, _("WARNING: %s: bad UUID\n"), s);
 	return NULL;
 }
 
@@ -371,3 +483,10 @@ char *interpret_spec(char *spec)
 		dev = string_copy(spec);
 	return dev;
 }
+
+#ifdef DEBUG
+main(int argc, char **argv)
+{
+	uuidcache_init();
+}
+#endif
