@@ -48,6 +48,7 @@ extern int optind;
 #include "e2p/e2p.h"
 #include "jfs_user.h"
 #include "util.h"
+#include "get_device_by_label.h"
 
 #include "../version.h"
 #include "nls-enable.h"
@@ -56,7 +57,8 @@ const char * program_name = "tune2fs";
 char * device_name;
 char * new_label, *new_last_mounted, *new_UUID;
 static int c_flag, C_flag, e_flag, f_flag, g_flag, i_flag, l_flag, L_flag;
-static int m_flag, M_flag, r_flag, s_flag = -1, u_flag, U_flag;
+static int m_flag, M_flag, r_flag, s_flag = -1, u_flag, U_flag, T_flag;
+static time_t last_check_time;
 static int print_label;
 static int max_mount_count, mount_count, mount_flags;
 static unsigned long interval, reserved_ratio, reserved_blocks;
@@ -64,27 +66,33 @@ static unsigned long resgid, resuid;
 static unsigned short errors;
 static int open_flag;
 static char *features_cmd;
+static char *mntopts_cmd;
 
 int journal_size, journal_flags;
 char *journal_device;
 
 static const char *please_fsck = N_("Please run e2fsck on the filesystem.\n");
 
+void do_findfs(int argc, char **argv);
+
 static void usage(void)
 {
 	fprintf(stderr,
 		_("Usage: %s [-c max-mounts-count] [-e errors-behavior] "
 		  "[-g group]\n"
-		 "\t[-i interval[d|m|w]] [-j] [-J journal-options]\n"
-		 "\t[-l] [-s sparse-flag] [-m reserved-blocks-percent]\n"
-		  "\t[-r reserved-blocks-count] [-u user] [-C mount-count]\n"
-		  "\t[-L volume-label] [-M last-mounted-dir] [-U UUID]\n"
-		  "\t[-O [^]feature[,...]] device\n"), program_name);
+		  "\t[-i interval[d|m|w]] [-j] [-J journal-options]\n"
+		  "\t[-l] [-s sparse-flag] [-m reserved-blocks-percent]\n"
+		  "\t[-o [^]mount-options[,...]] [-r reserved-blocks-count]\n"
+		  "\t[-u user] [-C mount-count] [-L volume-label] "
+		  "[-M last-mounted-dir]\n"
+		  "\t[-O [^]feature[,...]] [-T last-check-time] [-U UUID]"
+		  " device\n"), program_name);
 	exit (1);
 }
 
 static __u32 ok_features[3] = {
-	EXT3_FEATURE_COMPAT_HAS_JOURNAL,	/* Compat */
+	EXT3_FEATURE_COMPAT_HAS_JOURNAL |
+		EXT2_FEATURE_COMPAT_DIR_INDEX,	/* Compat */
 	EXT2_FEATURE_INCOMPAT_FILETYPE,		/* Incompat */
 	EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER	/* R/O compat */
 };
@@ -94,7 +102,7 @@ static __u32 ok_features[3] = {
  */
 static void remove_journal_device(ext2_filsys fs)
 {
-	char		*journal_device;
+	char		*journal_path;
 	ext2_filsys	jfs;
 	char		buf[1024];
 	journal_superblock_t	*jsb;
@@ -105,11 +113,17 @@ static void remove_journal_device(ext2_filsys fs)
 	if (f_flag)
 		commit_remove_journal = 1; /* force removal even if error */
 
-	journal_device = ext2fs_find_block_device(fs->super->s_journal_dev);
-	if (!journal_device)
-		return;
+	uuid_unparse(fs->super->s_journal_uuid, buf);
+	journal_path = get_spec_by_uuid(buf);
 
-	retval = ext2fs_open(journal_device, EXT2_FLAG_RW|
+	if (!journal_path) {
+		journal_path =
+			ext2fs_find_block_device(fs->super->s_journal_dev);
+		if (!journal_path)
+			return;
+	}
+
+	retval = ext2fs_open(journal_path, EXT2_FLAG_RW|
 			     EXT2_FLAG_JOURNAL_DEV_OK, 0,
 			     fs->blocksize, unix_io_manager, &jfs);
 	if (retval) {
@@ -118,8 +132,8 @@ static void remove_journal_device(ext2_filsys fs)
 		goto no_valid_journal;
 	}
 	if (!(jfs->super->s_feature_incompat & EXT3_FEATURE_INCOMPAT_JOURNAL_DEV)) {
-		fprintf(stderr, "%s is not a journal device.\n",
-			journal_device);
+		fprintf(stderr, _("%s is not a journal device.\n"),
+			journal_path);
 		goto no_valid_journal;
 	}
 
@@ -145,7 +159,8 @@ static void remove_journal_device(ext2_filsys fs)
 			break;
 	}
 	if (i >= nr_users) {
-		fprintf(stderr, "Filesystem's UUID not found on journal device.\n");
+		fprintf(stderr,
+			_("Filesystem's UUID not found on journal device.\n"));
 		commit_remove_journal = 1;
 		goto no_valid_journal;
 	}
@@ -165,7 +180,7 @@ static void remove_journal_device(ext2_filsys fs)
 
 no_valid_journal:
 	if (commit_remove_journal == 0) {
-		printf(_("Journal NOT removed\n"));
+		fprintf(stderr, _("Journal NOT removed\n"));
 		exit(1);
 	}
 	fs->super->s_journal_dev = 0;
@@ -173,8 +188,82 @@ no_valid_journal:
 	       sizeof(fs->super->s_journal_uuid));
 	ext2fs_mark_super_dirty(fs);
 	printf(_("Journal removed\n"));
+	free(journal_path);
 }
 
+/* Helper function for remove_journal_inode */
+static int release_blocks_proc(ext2_filsys fs, blk_t *blocknr,
+			       int blockcnt, void *private)
+{
+	blk_t	block;
+	int	group;
+
+	block = *blocknr;
+	ext2fs_unmark_block_bitmap(fs->block_map,block);
+	group = ext2fs_group_of_blk(fs, block);
+	fs->group_desc[group].bg_free_blocks_count++;
+	fs->super->s_free_blocks_count++;
+	return 0;
+}
+
+/*
+ * Remove the journal inode from the filesystem
+ */
+static void remove_journal_inode(ext2_filsys fs)
+{
+	struct ext2_inode	inode;
+	errcode_t		retval;
+	ino_t			ino = fs->super->s_journal_inum;
+	
+	retval = ext2fs_read_inode(fs, ino,  &inode);
+	if (retval) {
+		com_err(program_name, retval,
+			_("while reading journal inode"));
+		exit(1);
+	}
+	if (ino == EXT2_JOURNAL_INO) {
+		retval = ext2fs_read_bitmaps(fs);
+		if (retval) {
+			com_err(program_name, retval,
+				_("while reading bitmaps"));
+			exit(1);
+		}
+		retval = ext2fs_block_iterate(fs, ino, 0, NULL,
+					      release_blocks_proc, NULL);
+		if (retval) {
+			com_err(program_name, retval,
+				_("while clearing journal inode"));
+			exit(1);
+		}
+		memset(&inode, 0, sizeof(inode));
+		ext2fs_mark_bb_dirty(fs);
+		fs->flags &= ~EXT2_FLAG_SUPER_ONLY;
+	} else
+		inode.i_flags &= ~EXT2_IMMUTABLE_FL;
+	retval = ext2fs_write_inode(fs, ino, &inode);
+	if (retval) {
+		com_err(program_name, retval,
+			_("while writing journal inode"));
+		exit(1);
+	}
+	fs->super->s_journal_inum = 0;
+	ext2fs_mark_super_dirty(fs);
+}
+
+/*
+ * Update the default mount options
+ */
+static void update_mntopts(ext2_filsys fs, char *mntopts)
+{
+	struct ext2_super_block *sb= fs->super;
+
+	if (e2p_edit_mntopts(mntopts, &sb->s_default_mount_opts, ~0)) {
+		fprintf(stderr, _("Invalid mount option set: %s\n"),
+			mntopts);
+		exit(1);
+	}
+	ext2fs_mark_super_dirty(fs);
+}
 
 /*
  * Update the feature set as provided by the user.
@@ -182,10 +271,8 @@ no_valid_journal:
 static void update_feature_set(ext2_filsys fs, char *features)
 {
 	int sparse, old_sparse, filetype, old_filetype;
-	int journal, old_journal;
-	struct ext2_inode	inode;
+	int journal, old_journal, dxdir, old_dxdir;
 	struct ext2_super_block *sb= fs->super;
-	errcode_t		retval;
 
 	old_sparse = sb->s_feature_ro_compat &
 		EXT2_FEATURE_RO_COMPAT_SPARSE_SUPER;
@@ -193,6 +280,8 @@ static void update_feature_set(ext2_filsys fs, char *features)
 		EXT2_FEATURE_INCOMPAT_FILETYPE;
 	old_journal = sb->s_feature_compat &
 		EXT3_FEATURE_COMPAT_HAS_JOURNAL;
+	old_dxdir = sb->s_feature_compat &
+		EXT2_FEATURE_COMPAT_DIR_INDEX;
 	if (e2p_edit_feature(features, &sb->s_feature_compat,
 			     ok_features)) {
 		fprintf(stderr, _("Invalid filesystem option set: %s\n"),
@@ -205,6 +294,8 @@ static void update_feature_set(ext2_filsys fs, char *features)
 		EXT2_FEATURE_INCOMPAT_FILETYPE;
 	journal = sb->s_feature_compat &
 		EXT3_FEATURE_COMPAT_HAS_JOURNAL;
+	dxdir = sb->s_feature_compat &
+		EXT2_FEATURE_COMPAT_DIR_INDEX;
 	if (old_journal && !journal) {
 		if ((mount_flags & EXT2_MF_MOUNTED) &&
 		    !(mount_flags & EXT2_MF_READONLY)) {
@@ -223,28 +314,12 @@ static void update_feature_set(ext2_filsys fs, char *features)
 				  "the has_journal flag.\n"));
 			exit(1);
 		}
-		/*
-		 * Remove the immutable flag on the journal inode
-		 */
 		if (sb->s_journal_inum) {
-			retval = ext2fs_read_inode(fs, sb->s_journal_inum, 
-						   &inode);
-			if (retval) {
-				com_err(program_name, retval,
-					"while reading journal inode");
-				exit(1);
-			}
-			inode.i_flags &= ~EXT2_IMMUTABLE_FL;
-			retval = ext2fs_write_inode(fs, sb->s_journal_inum, 
-						    &inode);
-			if (retval) {
-				com_err(program_name, retval,
-					"while writing journal inode");
-				exit(1);
-			}
+			remove_journal_inode(fs);
 		}
-		if (sb->s_journal_dev)
+		if (sb->s_journal_dev) {
 			remove_journal_device(fs);
+		}
 	}
 	if (journal && !old_journal) {
 		/*
@@ -256,7 +331,12 @@ static void update_feature_set(ext2_filsys fs, char *features)
 		if (!journal_size)
 			journal_size = -1;
 		sb->s_feature_compat &= ~EXT3_FEATURE_COMPAT_HAS_JOURNAL;
-		journal = old_journal;
+	}
+	if (dxdir && !old_dxdir) {
+		if (!sb->s_def_hash_version)
+			sb->s_def_hash_version = EXT2_HASH_TEA;
+		if (uuid_is_null((unsigned char *) sb->s_hash_seed))
+			uuid_generate((unsigned char *) sb->s_hash_seed);
 	}
 
 	if (sb->s_rev_level == EXT2_GOOD_OLD_REV &&
@@ -264,8 +344,7 @@ static void update_feature_set(ext2_filsys fs, char *features)
 	     sb->s_feature_incompat))
 		ext2fs_update_dynamic_rev(fs);
 	if ((sparse != old_sparse) ||
-	    (filetype != old_filetype) ||
-	    (journal != old_journal)) {
+	    (filetype != old_filetype)) {
 		sb->s_state &= ~EXT2_VALID_FS;
 		printf("\n%s\n", _(please_fsck));
 	}
@@ -284,7 +363,7 @@ static void add_journal(ext2_filsys fs)
 	if (fs->super->s_feature_compat &
 	    EXT3_FEATURE_COMPAT_HAS_JOURNAL) {
 		fprintf(stderr, _("The filesystem already has a journal.\n"));
-		exit(1);
+		goto err;
 	}
 	if (journal_device) {
 		check_plausibility(journal_device);
@@ -296,21 +375,21 @@ static void add_journal(ext2_filsys fs)
 			com_err(program_name, retval,
 				_("\n\twhile trying to open journal on %s\n"),
 				journal_device);
-			exit(1);
+			goto err;
 		}
 		printf(_("Creating journal on device %s: "),
 		       journal_device);
 		fflush(stdout);
-		
+
 		retval = ext2fs_add_journal_device(fs, jfs);
+		ext2fs_close(jfs);
 		if (retval) {
 			com_err (program_name, retval,
 				 _("while adding filesystem to journal on %s"),
 				 journal_device);
-			exit(1);
+			goto err;
 		}
 		printf(_("done\n"));
-		ext2fs_close(jfs);
 	} else if (journal_size) {
 		printf(_("Creating journal inode: "));
 		fflush(stdout);
@@ -319,7 +398,7 @@ static void add_journal(ext2_filsys fs)
 		retval = ext2fs_add_journal_inode(fs, journal_blocks,
 						  journal_flags);
 		if (retval) {
-			printf("\n");
+			fprintf(stderr, "\n");
 			com_err(program_name, retval,
 				_("\n\twhile trying to create journal file"));
 			exit(1);
@@ -333,20 +412,12 @@ static void add_journal(ext2_filsys fs)
 			fs->flags &= ~EXT2_FLAG_SUPER_ONLY;
 	}
 	print_check_message(fs);
-}
+	return;
 
-/*
- * Given argv[0], return the program name.
- */
-static char *get_progname(char *argv_zero)
-{
-	char	*cp;
-
-	cp = strrchr(argv_zero, '/');
-	if (!cp )
-		return argv_zero;
-	else
-		return cp+1;
+err:
+	if (journal_device)
+		free(journal_device);
+	exit(1);
 }
 
 
@@ -358,13 +429,30 @@ static void parse_e2label_options(int argc, char ** argv)
 	}
 	device_name = argv[1];
 	if (argc == 3) {
-		open_flag = EXT2_FLAG_RW;
+		open_flag = EXT2_FLAG_RW | EXT2_FLAG_JOURNAL_DEV_OK;
 		L_flag = 1;
 		new_label = argv[2];
 	} else 
 		print_label++;
 }
 
+static time_t parse_time(char *str)
+{
+	struct	tm	ts;
+
+	if (strcmp(str, "now") == 0) {
+		return (time(0));
+	}
+	memset(&ts, 0, sizeof(ts));
+	strptime(optarg, "%Y%m%d%H%M%S", &ts);
+	if (ts.tm_mday == 0) {
+		com_err(program_name, 0,
+			_("Couldn't parse date/time specifier: %s"),
+			str);
+		usage();
+	}
+	return (mktime(&ts));
+}
 
 static void parse_tune2fs_options(int argc, char **argv)
 {
@@ -373,10 +461,8 @@ static void parse_tune2fs_options(int argc, char **argv)
 	struct group * gr;
 	struct passwd * pw;
 
-	fprintf (stderr, _("tune2fs %s, %s for EXT2 FS %s, %s\n"),
-		 E2FSPROGS_VERSION, E2FSPROGS_DATE,
-		 EXT2FS_VERSION, EXT2FS_DATE);
-	while ((c = getopt (argc, argv, "c:e:fg:i:jlm:r:s:u:C:J:L:M:O:U:")) != EOF)
+	printf("tune2fs %s (%s)\n", E2FSPROGS_VERSION, E2FSPROGS_DATE);
+	while ((c = getopt(argc, argv, "c:e:fg:i:jlm:o:r:s:u:C:J:L:M:O:T:U:")) != EOF)
 		switch (c)
 		{
 			case 'c':
@@ -489,7 +575,8 @@ static void parse_tune2fs_options(int argc, char **argv)
 			case 'L':
 				new_label = optarg;
 				L_flag = 1;
-				open_flag = EXT2_FLAG_RW;
+				open_flag = EXT2_FLAG_RW |
+					EXT2_FLAG_JOURNAL_DEV_OK;
 				break;
 			case 'm':
 				reserved_ratio = strtoul (optarg, &tmp, 0);
@@ -507,7 +594,22 @@ static void parse_tune2fs_options(int argc, char **argv)
 				M_flag = 1;
 				open_flag = EXT2_FLAG_RW;
 				break;
+			case 'o':
+				if (mntopts_cmd) {
+					com_err (program_name, 0,
+					 _("-o may only be specified once"));
+					usage();
+				}
+				mntopts_cmd = optarg;
+				open_flag = EXT2_FLAG_RW;
+				break;
+				
 			case 'O':
+				if (features_cmd) {
+					com_err (program_name, 0,
+					 _("-O may only be specified once"));
+					usage();
+				}
 				features_cmd = optarg;
 				open_flag = EXT2_FLAG_RW;
 				break;
@@ -526,6 +628,11 @@ static void parse_tune2fs_options(int argc, char **argv)
 				s_flag = atoi(optarg);
 				open_flag = EXT2_FLAG_RW;
 				break;
+			case 'T':
+				T_flag = 1;
+				last_check_time = parse_time(optarg);
+				open_flag = EXT2_FLAG_RW;
+				break;
 			case 'u':
 				resuid = strtoul (optarg, &tmp, 0);
 				if (*tmp) {
@@ -549,7 +656,8 @@ static void parse_tune2fs_options(int argc, char **argv)
 			case 'U':
 				new_UUID = optarg;
 				U_flag = 1;
-				open_flag = EXT2_FLAG_RW;
+				open_flag = EXT2_FLAG_RW |
+					EXT2_FLAG_JOURNAL_DEV_OK;
 				break;
 			default:
 				usage();
@@ -559,8 +667,26 @@ static void parse_tune2fs_options(int argc, char **argv)
 	if (!open_flag && !l_flag)
 		usage();
 	device_name = argv[optind];
-}	
+}
 
+void do_findfs(int argc, char **argv)
+{
+	char	*dev;
+
+	if ((argc != 2) ||
+	    (strncmp(argv[1], "LABEL=", 6) && strncmp(argv[1], "UUID=", 5))) {
+		fprintf(stderr, "Usage: findfs LABEL=<label>|UUID=<uuid>\n");
+		exit(2);
+	}
+	dev = interpret_spec(argv[1]);
+	if (!dev) {
+		fprintf(stderr, "Filesystem matching %s not found\n", 
+			argv[1]);
+		exit(1);
+	}
+	puts(dev);
+	exit(0);
+}
 
 
 int main (int argc, char ** argv)
@@ -571,6 +697,7 @@ int main (int argc, char ** argv)
 
 #ifdef ENABLE_NLS
 	setlocale(LC_MESSAGES, "");
+	setlocale(LC_CTYPE, "");
 	bindtextdomain(NLS_CAT_NAME, LOCALEDIR);
 	textdomain(NLS_CAT_NAME);
 #endif
@@ -578,6 +705,8 @@ int main (int argc, char ** argv)
 		program_name = *argv;
 	initialize_ext2_error_table();
 
+	if (strcmp(get_progname(argv[0]), "findfs") == 0)
+		do_findfs(argc, argv);
 	if (strcmp(get_progname(argv[0]), "e2label") == 0)
 		parse_e2label_options(argc, argv);
 	else
@@ -588,7 +717,8 @@ int main (int argc, char ** argv)
         if (retval) {
 		com_err (program_name, retval, _("while trying to open %s"),
 			 device_name);
-		printf(_("Couldn't find valid filesystem superblock.\n"));
+		fprintf(stderr,
+			_("Couldn't find valid filesystem superblock.\n"));
 		exit(1);
 	}
 	sb = fs->super;
@@ -682,6 +812,12 @@ int main (int argc, char ** argv)
 			       _(please_fsck));
 		}
 	}
+	if (T_flag) {
+		sb->s_lastcheck = last_check_time;
+		ext2fs_mark_super_dirty(fs);
+		printf(_("Setting time filesystem last checked to %s\n"),
+		       ctime(&last_check_time));
+	}
 	if (u_flag) {
 		sb->s_def_resuid = resuid;
 		ext2fs_mark_super_dirty(fs);
@@ -702,6 +838,8 @@ int main (int argc, char ** argv)
 			sizeof(sb->s_last_mounted));
 		ext2fs_mark_super_dirty(fs);
 	}
+	if (mntopts_cmd)
+		update_mntopts(fs, mntopts_cmd);
 	if (features_cmd)
 		update_feature_set(fs, features_cmd);
 	if (journal_size || journal_device)
@@ -724,6 +862,5 @@ int main (int argc, char ** argv)
 
 	if (l_flag)
 		list_super (sb);
-	ext2fs_close (fs);
-	exit (0);
+	return (ext2fs_close (fs) ? 1 : 0);
 }
