@@ -9,6 +9,9 @@
  * %End-Header%
  */
 
+#define _LARGEFILE_SOURCE
+#define _LARGEFILE64_SOURCE
+
 #include <stdio.h>
 #include <string.h>
 #ifdef HAVE_ERRNO_H
@@ -26,6 +29,7 @@
 #include "ext2fs/ext2_fs.h"
 #include "ext2fs/ext2fs.h"
 #include "nls-enable.h"
+#include "get_device_by_label.h"
 #include "util.h"
 
 #ifndef HAVE_STRCASECMP
@@ -44,6 +48,20 @@ int strcasecmp (char *s1, char *s2)
 }
 #endif
 
+/*
+ * Given argv[0], return the program name.
+ */
+char *get_progname(char *argv_zero)
+{
+	char	*cp;
+
+	cp = strrchr(argv_zero, '/');
+	if (!cp )
+		return argv_zero;
+	else
+		return cp+1;
+}
+
 void proceed_question(void)
 {
 	char buf[256];
@@ -61,9 +79,15 @@ void proceed_question(void)
 void check_plausibility(const char *device)
 {
 	int val;
+#ifdef HAVE_OPEN64
+	struct stat64 s;
+	
+	val = stat64(device, &s);
+#else
 	struct stat s;
 	
 	val = stat(device, &s);
+#endif
 	
 	if(val == -1) {
 		fprintf(stderr, _("Could not stat %s --- %s\n"),
@@ -154,11 +178,11 @@ void parse_journal_opts(const char *opts)
 		       arg ? arg : "NONE");
 #endif
 		if (strcmp(token, "device") == 0) {
-			if (!arg) {
+			journal_device = interpret_spec(arg);
+			if (!journal_device) {
 				journal_usage++;
 				continue;
 			}
-			journal_device = arg;
 		} else if (strcmp(token, "size") == 0) {
 			if (!arg) {
 				journal_usage++;
@@ -196,7 +220,7 @@ void parse_journal_opts(const char *opts)
  * in the filesystem.  For very small filesystems, it is not reasonable to
  * have a journal that fills more than half of the filesystem.
  */
-int figure_journal_size(int journal_size, ext2_filsys fs)
+int figure_journal_size(int size, ext2_filsys fs)
 {
 	blk_t j_blocks;
 
@@ -205,9 +229,8 @@ int figure_journal_size(int journal_size, ext2_filsys fs)
 		return 0;
 	}
 	
-	if (journal_size >= 0) {
-		j_blocks = journal_size * 1024 /
-			(fs->blocksize	/ 1024);
+	if (size >= 0) {
+		j_blocks = size * 1024 / (fs->blocksize	/ 1024);
 		if (j_blocks < 1024 || j_blocks > 102400) {
 			fprintf(stderr, _("\nThe requested journal "
 				"size is %d blocks; it must be\n"
