@@ -10,17 +10,24 @@
  */
 
 #include <stdio.h>
+#ifdef HAVE_STDLIB_H
+#include <stdlib.h>
+#endif
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
+#else
+extern char *optarg;
+extern int optind;
 #endif
 #include "uuid/uuid.h"
+#include "nls-enable.h"
 
 #define DO_TYPE_TIME	1
 #define DO_TYPE_RANDOM	2
 
-void usage(const char *progname)
+static void usage(const char *progname)
 {
-	fprintf(stderr, "Usage: %s [-r] [-t]\n", progname);
+	fprintf(stderr, _("Usage: %s [-r] [-t]\n"), progname);
 	exit(1);
 }
 
