@@ -19,6 +19,7 @@ extern char *optarg;
 extern int optind;
 #endif
 #include <fcntl.h>
+#include <sys/stat.h>
 
 #include "resize2fs.h"
 
@@ -119,6 +120,7 @@ int main (int argc, char ** argv)
 	blk_t		max_size = 0;
 	io_manager	io_ptr;
 	char		*tmp;
+	struct stat	st_buf;
 	
 	initialize_ext2_error_table();
 
@@ -202,8 +204,7 @@ int main (int argc, char ** argv)
 	 * Check for compatibility with the feature sets.  We need to
 	 * be more stringent than ext2fs_open().
 	 */
-	if ((fs->super->s_feature_compat & ~EXT2_LIB_FEATURE_COMPAT_SUPP) ||
-	    (fs->super->s_feature_incompat & ~EXT2_LIB_FEATURE_RO_COMPAT_SUPP)) {
+	if (fs->super->s_feature_compat & ~EXT2_LIB_FEATURE_COMPAT_SUPP) {
 		com_err(program_name, EXT2_ET_UNSUPP_FEATURE,
 			"(%s)", device_name);
 		exit(1);
@@ -223,6 +224,21 @@ int main (int argc, char ** argv)
 	}
 	if (!new_size)
 		new_size = max_size;
+	/*
+	 * If we are resizing a plain file, and it's not big enough,
+	 * automatically extend it in a sparse fashion by writing the
+	 * last requested block.
+	 */
+	if ((new_size > max_size) &&
+	    (stat(device_name, &st_buf) == 0) &&
+	    S_ISREG(st_buf.st_mode) &&
+	    ((tmp = malloc(fs->blocksize)) != 0)) {
+		memset(tmp, 0, fs->blocksize);
+		retval = io_channel_write_blk(fs->io, new_size-1, 1, tmp);
+		if (retval == 0)
+			max_size = new_size;
+		free(tmp);
+	}
 	if (!force && (new_size > max_size)) {
 		fprintf(stderr, _("The containing partition (or device)"
 			" is only %d blocks.\nYou requested a new size"
@@ -235,18 +251,21 @@ int main (int argc, char ** argv)
 			"long.  Nothing to do!\n\n"), new_size);
 		exit(0);
 	}
-	if (!force && (fs->super->s_lastcheck < fs->super->s_mtime)) {
+	if (!force && ((fs->super->s_lastcheck < fs->super->s_mtime) ||
+		       (fs->super->s_state & EXT2_ERROR_FS) ||
+		       ((fs->super->s_state & EXT2_VALID_FS) == 0))) {
 		fprintf(stderr, _("Please run 'e2fsck -f %s' first.\n\n"),
 			device_name);
 		exit(1);
 	}
-	retval = resize_fs(fs, new_size, flags,
+	retval = resize_fs(fs, &new_size, flags,
 			   ((flags & RESIZE_PERCENT_COMPLETE) ?
 			    resize_progress_func : 0));
 	if (retval) {
 		com_err(program_name, retval, _("while trying to resize %s"),
 			device_name);
 		ext2fs_close (fs);
+		exit(1);
 	}
 	printf(_("The filesystem on %s is now %d blocks long.\n\n"),
 	       device_name, new_size);
